"""Defines the wraps_ufunc decorator used to perform unit conversions and dimension handling."""
import functools
import warnings
from collections.abc import Callable, Mapping, Sequence, Set
from inspect import Parameter, Signature, signature
from types import GenericAlias
from typing import Any, Optional, Union

import xarray as xr
from pint import Unit, UnitStrippedWarning

from .setup_unit_handling import Quantity, convert_units, magnitude, ureg

FunctionType = Callable[..., Any]


def wraps_ufunc(  # noqa: PLR0915
    input_units: dict[str, Union[str, Unit, None]],
    return_units: dict[str, Union[str, Unit, None]],
    pass_as_kwargs: tuple = (),
    #  kwargs for apply_ufunc
    input_core_dims: Optional[Sequence[Sequence]] = None,
    output_core_dims: Optional[Sequence[Sequence]] = ((),),
    exclude_dims: Set = frozenset(),
    vectorize: bool = True,
    join: str = "exact",
    dataset_join: str = "exact",
    keep_attrs: str = "drop_conflicts",
    dask: str = "forbidden",
    output_dtypes: Optional[Sequence] = None,
    output_sizes: Optional[Mapping[Any, int]] = None,
    dask_gufunc_kwargs: Optional[dict[str, Any]] = None,
) -> FunctionType:
    """Decorator for functions to add in unit and dimension handling.

    input_units and return_units must be provided, as dictionaries giving
    a mapping between the function arguments/returns and their units.

    pass_as_kwargs can be used to optionally declare that specific arguments
    should be pass directly into the function, rather than vectorized.

    The remaining arguments for the wrapper correspond to arguments for
    xr.apply_ufunc.
    https://docs.xarray.dev/en/stable/examples/apply_ufunc_vectorize_1d.html
    """
    input_units = _check_units(input_units)
    return_units = _check_units(return_units)

    ufunc_kwargs: dict[str, Any] = dict(
        input_core_dims=input_core_dims,
        output_core_dims=output_core_dims,
        exclude_dims=exclude_dims,
        vectorize=vectorize,
        join=join,
        dataset_join=dataset_join,
        keep_attrs=keep_attrs,
        dask=dask,
        output_dtypes=output_dtypes,
        output_sizes=output_sizes,
        dask_gufunc_kwargs=dask_gufunc_kwargs,
    )
    input_keys = list(input_units.keys())

    if not isinstance(pass_as_kwargs, tuple):
        raise ValueError(f"pass_as_kwargs must be passed as a tuple of keys, not {str(type(pass_as_kwargs))[1:-1]}")

    pass_as_positional_args = [key for key in input_keys if key not in pass_as_kwargs]
    for arg in pass_as_kwargs:
        kwarg_position = input_keys.index(arg)
        if kwarg_position < len(pass_as_positional_args):
            raise ValueError(f"Argument {arg} in pass_as_kwargs appears before the positional args {pass_as_positional_args}")

    if input_core_dims is not None:
        if not len(input_core_dims) == len(pass_as_positional_args):
            raise ValueError(
                f"input_core_dims (len {len(input_core_dims)}) must the same length as positional_args ({pass_as_positional_args}, len {len(pass_as_positional_args)})"
            )
    else:
        input_core_dims = len(pass_as_positional_args) * [()]

    def _wraps_ufunc(func: FunctionType) -> FunctionType:

        func_signature = signature(func)
        func_parameters = func_signature.parameters

        if not list(input_units.keys()) == list(func_parameters.keys()):
            raise ValueError(
                f"Keys for input_units {input_units.keys()} did not match func_parameters {func_parameters.keys()} (n.b. order matters!)"
            )

        default_values = {key: val.default for key, val in func_parameters.items() if val.default is not Parameter.empty}

        @functools.wraps(func)
        def popcon_ufunc_wrapped_call(*args: Any, **kwargs: Any) -> Any:  # noqa: PLR0912
            """Transform args and kwargs, then call the inner function."""
            # if anything goes wrong we can do some extra work to provide a better error below
            try:
                args_dict = dict(zip(input_keys, args))

                if not set(args_dict.keys()).isdisjoint(kwargs.keys()):
                    raise RuntimeError(
                        f"{func.__name__} was called with repeat arguments. Input was interpreted as args={args_dict}, kwargs={kwargs}"
                    )

                args_dict = {**args_dict, **kwargs}
                args_dict = {**args_dict, **{key: val for key, val in default_values.items() if key not in args_dict.keys()}}

                args_dict = _return_magnitude_in_specified_units(args_dict, input_units)

                positional_args = []
                for i, key in enumerate(pass_as_positional_args):
                    arg = args_dict[key]
                    if not isinstance(arg, xr.DataArray):
                        positional_args.append(xr.DataArray(arg).expand_dims(input_core_dims[i]))
                    else:
                        positional_args.append(arg)

                with warnings.catch_warnings():
                    warnings.simplefilter("error", category=UnitStrippedWarning)
                    function_return = xr.apply_ufunc(
                        func,
                        *positional_args,
                        kwargs={key: args_dict[key] for key in pass_as_kwargs},
                        **ufunc_kwargs,
                    )

                if len(return_units) == 0:
                    # Assume that the function return None
                    return function_return.item()

                function_return = _convert_return_to_quantities(function_return, return_units)

                function_return = list(function_return.values())

                if len(function_return) > 1:
                    return tuple(function_return)
                else:
                    return function_return[0]

            except Exception as e:
                # the below checks if we are inside FunctionWrapper being called from another FunctionWrapper
                # if that is the case we try and give a more helpful error
                # if anything goes wrong in our frame inspection or we find that we aren't in a chained
                # call we raise the previous exception
                err = ""
                try:
                    import inspect

                    frames = inspect.getouterframes(inspect.currentframe())
                    # the first entry is the current call so check if any of the earlier callees are a __call__ from a FunctionWrapper
                    for frame in frames[1:]:
                        if frame.function == "popcon_ufunc_wrapped_call":
                            f = frames[1]
                            err = "Calling `wraps_ufunc` decorated function from within `wraps_ufunc` decorated function is not allowed!\n"
                            err += f"Error at {f.filename}:{f.lineno}\n"
                            err += "\n".join(f.code_context) if f.code_context else ""
                            err += f"Try using `{frames[0].frame.f_locals['func'].__name__}.unitless_func(...)` instead."
                            break
                except Exception:
                    # error while determining if we are withing a chained FunctionWrapper so re-raise original error
                    raise e from None

                # if err is not empty we have determined we are within a chained call so we raise a better error
                if err:
                    raise RuntimeError(err) from None
                else:
                    raise e

        # more meaningfull alias to the scalar non-unit version of the function
        popcon_ufunc_wrapped_call.unitless_func = popcon_ufunc_wrapped_call.__wrapped__  # type:ignore[attr-defined]
        popcon_ufunc_wrapped_call.__signature__ = _make_new_sig(func_signature, input_units, return_units)  # type:ignore[attr-defined]
        return popcon_ufunc_wrapped_call

    return _wraps_ufunc


def _check_units(units_dict: dict[str, Union[str, Unit, None]]) -> dict[str, Union[str, Unit, None]]:

    for key, unit in units_dict.items():
        if unit is None:
            pass
        elif isinstance(unit, str):
            units_dict[key] = ureg(unit).units
        elif not isinstance(unit, Unit):
            raise TypeError(f"wraps_ufunc units for {key} must by of type str or Unit, not {str(type(unit))[1:-1]} (value was {unit})")

    return units_dict


def _return_magnitude_in_specified_units(vals: Any, units_mapping: dict[str, Union[str, Unit, None]]) -> dict[str, Any]:

    if not set(vals.keys()) == set(units_mapping):
        raise ValueError(f"Argument keys {vals.keys()} did not match units_mapping keys {units_mapping.keys()}")

    converted_vals = {}

    for key in vals.keys():
        val = vals[key]
        unit = units_mapping[key]

        if unit is None or val is None:
            converted_vals[key] = val

        elif isinstance(val, Quantity):
            converted_vals[key] = magnitude(convert_units(val, unit))

        elif isinstance(val, xr.DataArray):
            converted_vals[key] = convert_units(val, unit).pint.dequantify()

        elif Quantity(1, unit).check(ureg.dimensionless):
            converted_vals[key] = val

        else:
            raise NotImplementedError(f"Cannot convert {key} of type {str(type(val))[1:-1]} to units {unit}")

    return converted_vals


def _convert_return_to_quantities(vals: Any, units_mapping: dict[str, Union[str, Unit, None]]) -> dict[str, Any]:

    if not isinstance(vals, tuple):
        vals = (vals,)

    if not len(vals) == len(units_mapping):
        raise ValueError(f"Number of returned values ({len(vals)}) did not match length of units_mapping ({len(units_mapping)})")
    vals = dict(zip(units_mapping.keys(), vals))

    converted_vals = {}

    for key in vals.keys():
        val = vals[key]
        unit = units_mapping[key]

        if unit is None or val is None:
            converted_vals[key] = val

        elif isinstance(val, xr.DataArray):
            converted_vals[key] = val.pint.quantify(unit, unit_registry=ureg)

        elif isinstance(val, Quantity):
            converted_vals[key] = val.to(unit)

        else:
            converted_vals[key] = Quantity(val, unit)

    return converted_vals


def _make_new_sig(
    sig: Signature,
    input_units: Mapping[str, Union[str, Unit, None]],
    return_units: Mapping[str, Union[str, Unit, None]],
) -> Signature:
    """Create a new signature for a wrapped function that replaces the plain floats/arrays with Quantity/DataArray."""
    parameters = list(sig.parameters.values())
    ret_annotation = sig.return_annotation

    # update parameter annotations
    new_parameters: list[Parameter] = []
    for param, unit in zip(parameters, input_units.values()):
        if unit is None:
            new_parameters.append(param)
        else:
            new_parameters.append(param.replace(annotation=Union[Quantity, xr.DataArray]))

    # update return annotation
    units_list = list(return_units.values())

    # extract the types from the tuple
    if isinstance(ret_annotation, GenericAlias) and ret_annotation.__origin__ == tuple:
        old_types: list[Any] = list(ret_annotation.__args__)
    elif ret_annotation == Parameter.empty:
        old_types = [Any for _ in range(len(units_list))]
    else:
        old_types = [ret_annotation]

    if len(old_types) != len(units_list):
        warnings.warn(
            (
                f"Return type annotation {ret_annotation} has {len(old_types)} return values"
                f", while the return_units: {return_units} specifies {len(return_units)} values"
            ),
            stacklevel=3,
        )
    ret_types = tuple(xr.DataArray if units_list[i] is not None else old_types[i] for i in range(len(units_list)))

    if len(ret_types) == 0:
        new_ret_ann: Union[type, None, GenericAlias] = None
    elif len(ret_types) == 1:
        new_ret_ann = ret_types[0]
    else:
        new_ret_ann = GenericAlias(tuple, ret_types)

    return sig.replace(parameters=new_parameters, return_annotation=new_ret_ann)



"""Run the two point model with a fixed parallel heat flux density reaching the target."""
from typing import Union

import xarray as xr

from ..formulas.scrape_off_layer_model import solve_two_point_model
from ..named_options import MomentumLossFunction
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "upstream_electron_temp",
    "target_electron_density",
    "target_electron_temp",
    "target_electron_flux",
    "SOL_power_loss_fraction",
]


def run_two_point_model_fixed_qpart(
    target_q_parallel: Unitfull,
    q_parallel: Unitfull,
    parallel_connection_length: Unitfull,
    average_electron_density: Unitfull,
    nesep_over_nebar: Unitfull,
    toroidal_flux_expansion: Unitfull,
    fuel_average_mass_number: Unitfull,
    kappa_e0: Unitfull,
    SOL_momentum_loss_function: Union[MomentumLossFunction, xr.DataArray],
    raise_error_if_not_converged: bool = False,
) -> dict[str, Unitfull]:
    """Run the two point model with a fixed parallel heat flux density reaching the target.

    Args:
        target_q_parallel: :term:`glossary link<target_q_parallel>`
        q_parallel: :term:`glossary link<q_parallel>`
        parallel_connection_length: :term:`glossary link<parallel_connection_length>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        nesep_over_nebar: :term:`glossary link<nesep_over_nebar>`
        toroidal_flux_expansion: :term:`glossary link<toroidal_flux_expansion>`
        fuel_average_mass_number: :term:`glossary link<fuel_average_mass_number>`
        kappa_e0: :term:`glossary link<kappa_e0>`
        SOL_momentum_loss_function: :term:`glossary link<SOL_momentum_loss_function>`
        raise_error_if_not_converged: Raise an error if solve does not converge

    Returns:
        :term:`upstream_electron_temp`, :term:`target_electron_density`, :term:`target_electron_temp`, :term:`target_electron_flux`, :term:`SOL_power_loss_fraction`,

    """
    SOL_power_loss_fraction = (1.0 - target_q_parallel / q_parallel).clip(min=0.0, max=1.0)

    (upstream_electron_temp, target_electron_density, target_electron_temp, target_electron_flux,) = solve_two_point_model(
        SOL_power_loss_fraction=SOL_power_loss_fraction,
        parallel_heat_flux_density=q_parallel,
        parallel_connection_length=parallel_connection_length,
        upstream_electron_density=nesep_over_nebar * average_electron_density,
        toroidal_flux_expansion=toroidal_flux_expansion,
        fuel_average_mass_number=fuel_average_mass_number,
        kappa_e0=kappa_e0,
        SOL_momentum_loss_function=SOL_momentum_loss_function,
        raise_error_if_not_converged=raise_error_if_not_converged,
    )

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


two_point_model_fixed_qpart = Algorithm(
    function=run_two_point_model_fixed_qpart,
    return_keys=RETURN_KEYS,
)



"""Ohmic and bootstrap plasma current, loop resistivity & voltage, and current relaxation time."""
from ..unit_handling import Unitfull, ureg, wraps_ufunc


def calc_f_shaping(inverse_aspect_ratio: Unitfull, areal_elongation: Unitfull, triangularity_psi95: Unitfull) -> Unitfull:
    """Calculate the shaping function.

    Equation A11 from ITER Physics Basis Ch. 1. Eqn. A-11 :cite:`editors_iter_1999`
    See following discussion for how this function is used.
    q_95 = 5 * minor_radius^2 * magnetic_field_on_axis / (R * plasma_current) f_shaping

    Args:
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`
        triangularity_psi95: [~] :term:`glossary link<triangularity_psi95>`

    Returns:
        :term:`f_shaping` [~]
    """
    return ((1.0 + areal_elongation**2.0 * (1.0 + 2.0 * triangularity_psi95**2.0 - 1.2 * triangularity_psi95**3.0)) / 2.0) * (
        (1.17 - 0.65 * inverse_aspect_ratio) / (1.0 - inverse_aspect_ratio**2.0) ** 2.0
    )


@wraps_ufunc(
    input_units=dict(
        magnetic_field_on_axis=ureg.T,
        major_radius=ureg.m,
        inverse_aspect_ratio=ureg.dimensionless,
        q_star=ureg.dimensionless,
        f_shaping=ureg.dimensionless,
    ),
    return_units=dict(plasma_current=ureg.MA),
)
def calc_plasma_current(
    magnetic_field_on_axis: float, major_radius: float, inverse_aspect_ratio: float, q_star: float, f_shaping: float
) -> float:
    """Calculate the plasma current in mega-amperes.

    Updated formula from ITER Physics Basis Ch. 1. :cite:`editors_iter_1999`

    Args:
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        q_star: [~] :term:`glossary link<q_star>`
        f_shaping: [~] :term:`glossary link<f_shaping>`

    Returns:
        :term:`plasma_current` [MA]
    """
    return float(5.0 * ((inverse_aspect_ratio * major_radius) ** 2.0) * (magnetic_field_on_axis / (q_star * major_radius)) * f_shaping)


@wraps_ufunc(
    input_units=dict(
        magnetic_field_on_axis=ureg.T,
        major_radius=ureg.m,
        inverse_aspect_ratio=ureg.dimensionless,
        plasma_current=ureg.MA,
        f_shaping=ureg.dimensionless,
    ),
    return_units=dict(q_star=ureg.dimensionless),
)
def calc_q_star(
    magnetic_field_on_axis: float, major_radius: float, inverse_aspect_ratio: float, plasma_current: float, f_shaping: float
) -> float:
    """Calculate an analytical estimate for the edge safety factor q_star.

    Updated formula from ITER Physics Basis Ch. 1. :cite:`editors_iter_1999`

    Args:
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        f_shaping: [~] :term:`glossary link<f_shaping>`

    Returns:
        :term:`plasma_current` [MA]
    """
    return float(5.0 * (inverse_aspect_ratio * major_radius) ** 2.0 * magnetic_field_on_axis / (plasma_current * major_radius) * f_shaping)


def calc_ohmic_power(inductive_plasma_current: Unitfull, loop_voltage: Unitfull) -> Unitfull:
    """Calculate the Ohmic heating power.

    Args:
        inductive_plasma_current: [MA] :term:`glossary link<inductive_plasma_current>`
        loop_voltage: [V] :term:`glossary link<loop_voltage>`

    Returns:
        :term:`P_Ohmic` [MW]
    """
    return inductive_plasma_current * loop_voltage


@wraps_ufunc(input_units=dict(average_electron_temp=ureg.keV), return_units=dict(spitzer_resistivity=ureg.ohm * ureg.m))
def calc_Spitzer_loop_resistivity(average_electron_temp: float) -> float:
    """Calculate the parallel Spitzer loop resistivity assuming the Coulomb logarithm = 17 and Z=1.

    Resistivity from Wesson 2.16.2 :cite:`wesson_tokamaks_2011`

    Args:
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`

    Returns:
        :term:`spitzer_resistivity` [Ohm-m]
    """
    return float((2.8e-8) * (average_electron_temp ** (-1.5)))


def calc_resistivity_trapped_enhancement(inverse_aspect_ratio: Unitfull, definition: int = 3) -> Unitfull:
    """Calculate the enhancement of the plasma resistivity due to trapped particles.

    Definition 1 is the denominator of eta_n (neoclassical resistivity) on p801 of Wesson :cite:`wesson_tokamaks_2011`

    Args:
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        definition: [~] choice of [1,2,3] to specify which definition to use

    Returns:
        :term:`trapped_particle_fraction` [~]

    Raises:
        NotImplementedError: if definition doesn't match an implementation
    """
    if definition == 1:
        trapped_particle_fraction = 1 / ((1.0 - (inverse_aspect_ratio**0.5)) ** 2.0)  # pragma: nocover
    elif definition == 2:
        trapped_particle_fraction = 2 / (1.0 - 1.31 * (inverse_aspect_ratio**0.5) + 0.46 * inverse_aspect_ratio)  # pragma: nocover
    elif definition == 3:
        trapped_particle_fraction = 0.609 / (0.609 - 0.785 * (inverse_aspect_ratio**0.5) + 0.269 * inverse_aspect_ratio)
    else:
        raise NotImplementedError(f"No implementation {definition} for calc_resistivity_trapped_enhancement.")  # pragma: nocover

    return trapped_particle_fraction


def calc_neoclassical_loop_resistivity(
    spitzer_resistivity: Unitfull, z_effective: Unitfull, trapped_particle_fraction: Unitfull
) -> Unitfull:
    """Calculate the neoclassical loop resistivity including impurity ions.

    Wesson Section 14.10. Impact of ion charge. Impact of dilution ~ 0.9.

    Args:
        spitzer_resistivity: [Ohm-m] :term:`glossary link<spitzer_resistivity>`
        z_effective: [~] :term:`glossary link<z_effective>`
        trapped_particle_fraction: [~] :term:`glossary link<trapped_particle_fraction>`

    Returns:
        :term:`neoclassical_loop_resistivity` [Ohm-m]
    """
    return spitzer_resistivity * z_effective * 0.9 * trapped_particle_fraction


@wraps_ufunc(
    input_units=dict(
        major_radius=ureg.m,
        inverse_aspect_ratio=ureg.dimensionless,
        areal_elongation=ureg.dimensionless,
        average_electron_temp=ureg.keV,
        z_effective=ureg.dimensionless,
    ),
    return_units=dict(current_relaxation_time=ureg.s),
)
def calc_current_relaxation_time(
    major_radius: float, inverse_aspect_ratio: float, areal_elongation: float, average_electron_temp: float, z_effective: float
) -> float:
    """Calculate the current relaxation time.

    from :cite:`Bonoli`

    Args:
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        z_effective: [~] :term:`glossary link<z_effective>`

    Returns:
        :term:`current_relaxation_time` [s]
    """
    return float(
        1.4 * ((major_radius * inverse_aspect_ratio) ** 2.0) * areal_elongation * (average_electron_temp**1.5) / z_effective
    )  # [s]


def calc_loop_voltage(
    major_radius: Unitfull,
    minor_radius: Unitfull,
    inductive_plasma_current: Unitfull,
    areal_elongation: Unitfull,
    neoclassical_loop_resistivity: Unitfull,
) -> Unitfull:
    """Calculate plasma toroidal loop voltage at flattop.

    Plasma loop voltage from Alex Creely's original work.

    Args:
        major_radius: [m] :term:`glossary link<major_radius>`
        minor_radius: [m] :term:`glossary link<minor_radius>`
        inductive_plasma_current: [MA] :term:`glossary link<inductive_plasma_current>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`
        neoclassical_loop_resistivity: [Ohm-m] :term:`glossary link<neoclassical_loop_resistivity>`

    Returns:
        :term:`loop_voltage` [V]
    """
    Iind = inductive_plasma_current  # Inductive plasma current [A]

    _term1 = 2 * major_radius / (minor_radius**2 * areal_elongation)  # Toroidal length over plasma cross-section surface area [1/m]
    return Iind * _term1 * neoclassical_loop_resistivity


def calc_bootstrap_fraction(
    ion_density_peaking: Unitfull,
    electron_density_peaking: Unitfull,
    temperature_peaking: Unitfull,
    z_effective: Unitfull,
    q_star: Unitfull,
    inverse_aspect_ratio: Unitfull,
    beta_poloidal: Unitfull,
) -> Unitfull:
    """Calculate bootstrap current fraction.

    K. Gi et al, Bootstrap current fraction scaling :cite:`gi_bootstrap_2014`
    Equation assumes q0 = 1

    Args:
        ion_density_peaking: [~] :term:`glossary link<ion_density_peaking>`
        electron_density_peaking: [~] :term:`glossary link<electron_density_peaking>`
        temperature_peaking: [~] :term:`glossary link<temperature_peaking>`
        z_effective: [~] :term:`glossary link<z_effective>`
        q_star: [~] :term:`glossary link<q_star>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        beta_poloidal: [~] :term:`glossary link<beta_poloidal>`

    Returns:
        :term:`bootstrap_fraction` [~]
    """
    nu_n = (ion_density_peaking + electron_density_peaking) / 2

    bootstrap_fraction = 0.474 * (
        (temperature_peaking - 1.0 + nu_n - 1.0) ** 0.974
        * (temperature_peaking - 1.0) ** -0.416
        * z_effective**0.178
        * q_star**-0.133
        * inverse_aspect_ratio**0.4
        * beta_poloidal
    )

    return bootstrap_fraction



"""Operational limits to avoid disruptive regions."""
import numpy as np

from ..unit_handling import ureg, wraps_ufunc


@wraps_ufunc(
    return_units=dict(greenwald_fraction=ureg.dimensionless),
    input_units=dict(
        average_electron_density=ureg.n20, inverse_aspect_ratio=ureg.dimensionless, major_radius=ureg.m, plasma_current=ureg.MA
    ),
)
def calc_greenwald_fraction(
    average_electron_density: float, inverse_aspect_ratio: float, major_radius: float, plasma_current: float
) -> float:
    """Calculate the fraction of the Greenwald density limit.

    Args:
        average_electron_density: [1e20 m^-3] :term:`glossary link<average_electron_density>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        major_radius: [m] :term:`glossary link<major_radius>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`

    Returns:
        :term:`greenwald_fraction` [~]
    """
    n_Greenwald = calc_greenwald_density_limit.__wrapped__(plasma_current, inverse_aspect_ratio * major_radius)

    return float(average_electron_density / n_Greenwald)


@wraps_ufunc(return_units=dict(nG=ureg.n20), input_units=dict(plasma_current=ureg.MA, minor_radius=ureg.m))
def calc_greenwald_density_limit(plasma_current: float, minor_radius: float) -> float:
    """Calculate the Greenwald density limit.

    Args:
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        minor_radius: [m] :term:`glossary link<minor_radius>`

    Returns:
        nG Greenwald density limit [n20]
    """
    return plasma_current / (np.pi * minor_radius**2)


@wraps_ufunc(
    return_units=dict(troyon_max_beta=ureg.percent),
    input_units=dict(minor_radius=ureg.m, magnetic_field_on_axis=ureg.T, plasma_current=ureg.MA),
)
def calc_troyon_limit(minor_radius: float, magnetic_field_on_axis: float, plasma_current: float) -> float:
    """Calculate the maximum value for beta, according to the Troyon limit.

    Args:
        minor_radius: [m] :term:`glossary link<minor_radius>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`

    Returns:
         troyon_max_beta [~]
    """
    return 2.8 * plasma_current / (minor_radius * magnetic_field_on_axis)



"""OD figures-of-merit to characterize a design point."""
import numpy as np

from ..unit_handling import Unitfull, convert_units, ureg, wraps_ufunc


def calc_triple_product(peak_fuel_ion_density: Unitfull, peak_ion_temp: Unitfull, energy_confinement_time: Unitfull) -> Unitfull:
    """Calculate the fusion triple product.

    Args:
        peak_fuel_ion_density: [1e20 m^-3] :term:`glossary link<peak_fuel_ion_density>`
        peak_ion_temp: [keV] :term:`glossary link<peak_fuel_ion_density>`
        energy_confinement_time: [s] :term:`glossary link<energy_confinement_time>`

    Returns:
         fusion_triple_product [10e20 m**-3 keV s]
    """
    return peak_fuel_ion_density * peak_ion_temp * energy_confinement_time


def calc_rho_star(
    fuel_average_mass_number: Unitfull, average_ion_temp: Unitfull, magnetic_field_on_axis: Unitfull, minor_radius: Unitfull
) -> Unitfull:
    """Calculate rho* (normalized gyroradius).

    Equation 1a from :cite:`Verdoolaege_2021`

    Args:
        fuel_average_mass_number: [amu] :term:`glossary link<fuel_average_mass_number>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        minor_radius: [m] :term:`glossary link<minor_radius>`

    Returns:
         rho_star [~]
    """
    return convert_units(
        np.sqrt(fuel_average_mass_number * average_ion_temp) / (ureg.e * magnetic_field_on_axis * minor_radius), ureg.dimensionless
    )


@wraps_ufunc(input_units=dict(ne=ureg.m**-3, Te=ureg.eV), return_units=dict(Lambda_c=ureg.dimensionless))
def calc_coulomb_logarithm(ne: float, Te: float) -> float:
    """Calculate the Coulomb logarithm, for electron-electron or electron-ion collisions.

    From text on page 6 of :cite:`Verdoolaege_2021`
    """
    return float(30.9 - np.log(ne**0.5 * Te**-1.0))


def calc_normalised_collisionality(
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    average_ion_temp: Unitfull,
    q_star: Unitfull,
    major_radius: Unitfull,
    inverse_aspect_ratio: Unitfull,
    z_effective: Unitfull,
) -> Unitfull:
    """Calculate normalized collisionality.

    Equation 1c from :cite:`Verdoolaege_2021`

    Extra factor of ureg.e**2, presumably related to Te**-2 for Te in eV

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        q_star: [~] :term:`glossary link<q_star>`
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [m] :term:`glossary link<inverse_aspect_ratio>`
        z_effective: [~] :term:`glossary link<z_effective>`

    Returns:
         nu_star [~]
    """
    return convert_units(
        ureg.e**4
        / (2.0 * np.pi * 3**1.5 * ureg.epsilon_0**2)
        * calc_coulomb_logarithm(average_electron_density, average_electron_temp)
        * average_electron_density
        * q_star
        * major_radius
        * z_effective
        / (average_ion_temp**2 * inverse_aspect_ratio**1.5),
        ureg.dimensionless,
    )


def calc_peak_pressure(
    peak_electron_temp: Unitfull,
    peak_ion_temp: Unitfull,
    peak_electron_density: Unitfull,
    peak_fuel_ion_density: Unitfull,
) -> Unitfull:
    """Calculate the peak pressure (needed for solving for the magnetic equilibrium).

    Args:
        peak_electron_temp: [keV] :term:`glossary link<peak_electron_density>`
        peak_ion_temp: [keV] :term:`glossary link<peak_ion_temp>`
        peak_electron_density: [1e19 m^-3] :term:`glossary link<peak_electron_density>`
        peak_fuel_ion_density: [~] :term:`glossary link<peak_fuel_ion_density>`

    Returns:
         peak_pressure [Pa]
    """
    return convert_units(peak_electron_temp * peak_electron_density + peak_ion_temp * peak_fuel_ion_density, ureg.Pa)



"""Threshold powers required to enter improved confinement regimes."""
from ..named_options import ConfinementScaling
from ..unit_handling import ureg, wraps_ufunc


def _calc_Martin_LH_threshold(
    magnetic_field_on_axis: float, surface_area: float, fuel_average_mass_number: float, electron_density_profile: float
) -> float:
    """See below."""
    _DEUTERIUM_MASS_NUMBER = 2.0

    return float(0.0488 * ((electron_density_profile / 10.0) ** 0.717) * (magnetic_field_on_axis**0.803) * (surface_area**0.941)) * (
        _DEUTERIUM_MASS_NUMBER / fuel_average_mass_number
    )


@wraps_ufunc(
    return_units=dict(P_LH_thresh=ureg.MW),
    input_units=dict(
        plasma_current=ureg.MA,
        magnetic_field_on_axis=ureg.T,
        minor_radius=ureg.m,
        major_radius=ureg.m,
        surface_area=ureg.m**2,
        fuel_average_mass_number=ureg.amu,
        average_electron_density=ureg.n19,
        scale=ureg.dimensionless,
    ),
)
def calc_LH_transition_threshold_power(
    plasma_current: float,
    magnetic_field_on_axis: float,
    minor_radius: float,
    major_radius: float,
    surface_area: float,
    fuel_average_mass_number: float,
    average_electron_density: float,
    scale: float = 1.0,
) -> float:
    """Calculate the threshold power (crossing the separatrix) to transition into H-mode.

    From Martin NF 2008 Scaling, with mass correction :cite:`martin_power_2008`
    Added in low density branch from Ryter 2014 :cite:`Ryter_2014`

    Args:
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        minor_radius: [m] :term:`glossary link<minor_radius>`
        major_radius: [m] :term:`glossary link<major_radius>`
        surface_area: [m^2] :term:`glossary link<surface_area>`
        fuel_average_mass_number: [amu] :term:`glossary link<fuel_average_mass_number>`
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        scale: (optional) scaling factor for P_LH adjustment studies [~]

    Returns:
         :term:`P_LH_thresh` [MW]
    """
    neMin19 = (
        0.7 * (plasma_current**0.34) * (magnetic_field_on_axis**0.62) * (minor_radius**-0.95) * ((major_radius / minor_radius) ** 0.4)
    )

    if average_electron_density < neMin19:
        P_LH_thresh = _calc_Martin_LH_threshold(
            magnetic_field_on_axis, surface_area, fuel_average_mass_number, electron_density_profile=neMin19
        )
        return float(P_LH_thresh * (neMin19 / average_electron_density) ** 2.0) * scale
    else:
        P_LH_thresh = _calc_Martin_LH_threshold(
            magnetic_field_on_axis, surface_area, fuel_average_mass_number, electron_density_profile=average_electron_density
        )
        return P_LH_thresh * scale


@wraps_ufunc(
    return_units=dict(P_LI_thresh=ureg.MW),
    input_units=dict(plasma_current=ureg.MA, average_electron_density=ureg.n19, scale=ureg.dimensionless),
)
def calc_LI_transition_threshold_power(plasma_current: float, average_electron_density: float, scale: float = 1.0) -> float:
    """Calculate the threshold power (crossing the separatrix) to transition into I-mode.

    Note: uses scaling described in Fig 5 of ref :cite:`hubbard_threshold_2012`

    Args:
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        scale: (optional) scaling factor for P_LI adjustment studies [~]

    Returns:
        :term:`P_LI_thresh` [MW]
    """
    return float(2.11 * plasma_current**0.94 * ((average_electron_density / 10.0) ** 0.65)) * scale


@wraps_ufunc(
    return_units=dict(P_LH_thresh=ureg.MW),
    input_units=dict(
        energy_confinement_scaling=None,
        plasma_current=ureg.MA,
        magnetic_field_on_axis=ureg.T,
        minor_radius=ureg.m,
        major_radius=ureg.m,
        surface_area=ureg.m**2,
        fuel_average_mass_number=ureg.amu,
        average_electron_density=ureg.n19,
        confinement_threshold_scalar=ureg.dimensionless,
    ),
)
def calc_confinement_transition_threshold_power(
    energy_confinement_scaling: ConfinementScaling,
    plasma_current: float,
    magnetic_field_on_axis: float,
    minor_radius: float,
    major_radius: float,
    surface_area: float,
    fuel_average_mass_number: float,
    average_electron_density: float,
    confinement_threshold_scalar: float = 1.0,
) -> float:
    """Calculate the threshold power (crossing the separatrix) to transition into an improved confinement mode.

    Args:
        energy_confinement_scaling: [] :term:`glossary link<energy_confinement_scaling>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        minor_radius: [m] :term:`glossary link<minor_radius>`
        major_radius: [m] :term:`glossary link<major_radius>`
        surface_area: [m^2] :term:`glossary link<surface_area>`
        fuel_average_mass_number: [amu] :term:`glossary link<fuel_average_mass_number>`
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        confinement_threshold_scalar: (optional) scaling factor for P_LH adjustment studies [~]

    Returns:
        :term:`P_LH_thresh` [MW]
    """
    if energy_confinement_scaling not in [ConfinementScaling.LOC, ConfinementScaling.IModey2]:
        P_LH_thresh = calc_LH_transition_threshold_power.__wrapped__(
            plasma_current,
            magnetic_field_on_axis,
            minor_radius,
            major_radius,
            surface_area,
            fuel_average_mass_number,
            average_electron_density,
            scale=confinement_threshold_scalar,
        )
    elif energy_confinement_scaling == ConfinementScaling.IModey2:
        P_LH_thresh = calc_LI_transition_threshold_power.__wrapped__(
            plasma_current, average_electron_density, scale=confinement_threshold_scalar
        )
    else:
        raise ValueError("Encountered unhandled confinement time scaling.")

    return float(P_LH_thresh)



"""POPCON algorithms."""
from typing import Union

from ..named_options import Algorithms
from .algorithm_class import Algorithm, CompositeAlgorithm
from .beta import calc_beta
from .composite_algorithm import predictive_popcon
from .core_radiated_power import calc_core_radiated_power
from .extrinsic_core_radiator import calc_extrinsic_core_radiator
from .fusion_gain import calc_fusion_gain
from .geometry import calc_geometry
from .heat_exhaust import calc_heat_exhaust
from .ohmic_power import calc_ohmic_power
from .peaked_profiles import calc_peaked_profiles
from .plasma_current_from_q_star import calc_plasma_current_from_q_star
from .power_balance_from_tau_e import calc_power_balance_from_tau_e
from .q_star_from_plasma_current import calc_q_star_from_plasma_current
from .single_functions import SINGLE_FUNCTIONS
from .two_point_model_fixed_fpow import two_point_model_fixed_fpow
from .two_point_model_fixed_qpart import two_point_model_fixed_qpart
from .two_point_model_fixed_tet import two_point_model_fixed_tet
from .use_LOC_tau_e_below_threshold import use_LOC_tau_e_below_threshold
from .zeff_and_dilution_from_impurities import calc_zeff_and_dilution_from_impurities

ALGORITHMS: dict[Algorithms, Union[Algorithm, CompositeAlgorithm]] = {
    Algorithms["calc_beta"]: calc_beta,
    Algorithms["calc_core_radiated_power"]: calc_core_radiated_power,
    Algorithms["calc_extrinsic_core_radiator"]: calc_extrinsic_core_radiator,
    Algorithms["calc_fusion_gain"]: calc_fusion_gain,
    Algorithms["calc_geometry"]: calc_geometry,
    Algorithms["calc_heat_exhaust"]: calc_heat_exhaust,
    Algorithms["calc_ohmic_power"]: calc_ohmic_power,
    Algorithms["calc_peaked_profiles"]: calc_peaked_profiles,
    Algorithms["calc_plasma_current_from_q_star"]: calc_plasma_current_from_q_star,
    Algorithms["calc_power_balance_from_tau_e"]: calc_power_balance_from_tau_e,
    Algorithms["predictive_popcon"]: predictive_popcon,
    Algorithms["calc_q_star_from_plasma_current"]: calc_q_star_from_plasma_current,
    Algorithms["two_point_model_fixed_fpow"]: two_point_model_fixed_fpow,
    Algorithms["two_point_model_fixed_qpart"]: two_point_model_fixed_qpart,
    Algorithms["two_point_model_fixed_tet"]: two_point_model_fixed_tet,
    Algorithms["calc_zeff_and_dilution_from_impurities"]: calc_zeff_and_dilution_from_impurities,
    Algorithms["use_LOC_tau_e_below_threshold"]: use_LOC_tau_e_below_threshold,
    **SINGLE_FUNCTIONS,
}


def get_algorithm(algorithm: Union[Algorithms, str]) -> Union[Algorithm, CompositeAlgorithm]:
    """Accessor for algorithms."""
    if isinstance(algorithm, str):
        algorithm = Algorithms[algorithm]

    return ALGORITHMS[algorithm]


__all__ = [
    "calc_beta",
    "calc_core_radiated_power",
    "calc_extrinsic_core_radiator",
    "calc_fusion_gain",
    "calc_geometry",
    "calc_heat_exhaust",
    "calc_ohmic_power",
    "calc_peaked_profiles",
    "calc_plasma_current_from_q_star",
    "calc_power_balance_from_tau_e",
    "predictive_popcon",
    "calc_q_star_from_plasma_current",
    "two_point_model_fixed_fpow",
    "two_point_model_fixed_qpart",
    "two_point_model_fixed_tet",
    "calc_zeff_and_dilution_from_impurities",
    "ALGORITHMS",
    "get_algorithm",
]



"""Plasma geometry (inside the last-closed-flux-surface)."""
import numpy as np

from ..unit_handling import Unitfull


def calc_plasma_volume(major_radius: Unitfull, inverse_aspect_ratio: Unitfull, areal_elongation: Unitfull) -> Unitfull:
    """Calculate the plasma volume inside the last-closed-flux-surface.

    Args:
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`

    Returns:
         Vp [m^3]
    """
    return (
        2.0
        * np.pi
        * major_radius**3.0
        * inverse_aspect_ratio**2.0
        * areal_elongation
        * (np.pi - (np.pi - 8.0 / 3.0) * inverse_aspect_ratio)
    )


def calc_plasma_surface_area(major_radius: Unitfull, inverse_aspect_ratio: Unitfull, areal_elongation: Unitfull) -> Unitfull:
    """Calculate the plasma surface area inside the last-closed-flux-surface.

    Args:
        major_radius: [m] :term:`glossary link<major_radius>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`

    Returns:
         Sp [m^2]
    """
    return (
        2.0 * np.pi * (major_radius**2.0) * inverse_aspect_ratio * areal_elongation * (np.pi + 2.0 - (np.pi - 2.0) * inverse_aspect_ratio)
    )



"""Adds a readout of the field at the current mouse position for a colormapped field plotted with pcolormesh, contour, quiver, etc.

Usage:
>>> fig, ax = plt.subplots()
>>> ax.format_coord = CoordinateFormatter(x, y, z)
"""
import xarray as xr


class CoordinateFormatter:
    """Data storage object used for providing a coordinate formatter."""

    def __init__(self, array: xr.DataArray):  # pragma: nocover
        """Stores the data required for grid lookup."""
        self.array = array

    def __call__(self, mouse_x, mouse_y):  # pragma: nocover
        """Returns a string which gives the field value at the queried mouse position."""
        lookup = dict(zip(self.array.dims, (mouse_y, mouse_x)))

        mouse_z = float(self.array.sel(lookup, method="nearest").item())

        return f"x={mouse_x:f}, y={mouse_y:f}, z={mouse_z:f}"



"""Calculate the radiated power due to fuel and impurity species."""
from .inherent import calc_bremsstrahlung_radiation, calc_synchrotron_radiation
from .mavrin_coronal import calc_impurity_radiated_power_mavrin_coronal
from .mavrin_noncoronal import calc_impurity_radiated_power_mavrin_noncoronal
from .post_and_jensen import calc_impurity_radiated_power_post_and_jensen
from .radas import calc_impurity_radiated_power_radas
from .radiated_power import calc_impurity_radiated_power

__all__ = [
    "calc_bremsstrahlung_radiation",
    "calc_synchrotron_radiation",
    "calc_impurity_radiated_power_mavrin_coronal",
    "calc_impurity_radiated_power_mavrin_noncoronal",
    "calc_impurity_radiated_power_post_and_jensen",
    "calc_impurity_radiated_power_radas",
    "calc_impurity_radiated_power",
]



"""Module to perform simple scrape-off-layer calculations.

These are mostly based on the two-point-model, from :cite:`stangeby_2018`.
"""
from .lambda_q import calc_lambda_q
from .parallel_heat_flux_density import calc_parallel_heat_flux_density
from .solve_target_first_two_point_model import solve_target_first_two_point_model
from .solve_two_point_model import solve_two_point_model

__all__ = [
    "solve_two_point_model",
    "calc_parallel_heat_flux_density",
    "calc_lambda_q",
    "solve_target_first_two_point_model",
]



"""Compute all terms in the two-point-model for a fixed target electron temperature."""
from typing import Union

import xarray as xr

from ...named_options import MomentumLossFunction
from ...unit_handling import Unitfull, ureg
from .momentum_loss_functions import calc_SOL_momentum_loss_fraction
from .required_power_loss_fraction import calc_required_SOL_power_loss_fraction
from .target_electron_density import (
    calc_f_other_target_electron_density,
    calc_f_vol_loss_target_electron_density,
    calc_target_electron_density,
    calc_target_electron_density_basic,
)
from .target_electron_flux import (
    calc_f_other_target_electron_flux,
    calc_f_vol_loss_target_electron_flux,
    calc_target_electron_flux,
    calc_target_electron_flux_basic,
)
from .target_electron_temp import calc_f_other_target_electron_temp, calc_target_electron_temp_basic
from .total_pressure import calc_upstream_total_pressure
from .upstream_electron_temp import calc_upstream_electron_temp


def solve_target_first_two_point_model(
    target_electron_temp: Unitfull,
    parallel_heat_flux_density: Unitfull,
    parallel_connection_length: Unitfull,
    upstream_electron_density: Unitfull,
    toroidal_flux_expansion: Unitfull,
    fuel_average_mass_number: Unitfull,
    kappa_e0: Unitfull,
    SOL_momentum_loss_function: Union[MomentumLossFunction, xr.DataArray],
    sheath_heat_transmission_factor: Unitfull = 7.5 * ureg.dimensionless,
    SOL_conduction_fraction: Unitfull = 1.0 * ureg.dimensionless,
    target_ratio_of_ion_to_electron_temp: Unitfull = 1.0 * ureg.dimensionless,
    target_ratio_of_electron_to_ion_density: Unitfull = 1.0 * ureg.dimensionless,
    target_mach_number: Unitfull = 1.0 * ureg.dimensionless,
    upstream_ratio_of_ion_to_electron_temp: Unitfull = 1.0 * ureg.dimensionless,
    upstream_ratio_of_electron_to_ion_density: Unitfull = 1.0 * ureg.dimensionless,
    upstream_mach_number: Unitfull = 0.0 * ureg.dimensionless,
) -> tuple[Unitfull, Unitfull, Unitfull, Unitfull]:
    """Calculate the SOL_power_loss_fraction required to keep the target temperature at a given value.

    Args:
        target_electron_temp: [eV]
        parallel_heat_flux_density: [GW/m^2]
        parallel_connection_length: [m]
        upstream_electron_density: [m^-3]
        toroidal_flux_expansion: [~]
        fuel_average_mass_number: [~]
        kappa_e0: electron heat conductivity constant [W / (eV^3.5 * m)]
        SOL_momentum_loss_function: which momentum loss function to use
        sheath_heat_transmission_factor: [~]
        SOL_conduction_fraction: [~]
        target_ratio_of_ion_to_electron_temp: [~]
        target_ratio_of_electron_to_ion_density: [~]
        target_mach_number: [~]
        upstream_ratio_of_ion_to_electron_temp: [~]
        upstream_ratio_of_electron_to_ion_density: [~]
        upstream_mach_number: [~]

    Returns:
        SOL_power_loss_fraction [~], upstream_electron_temp [eV], target_electron_density [m^-3], target_electron_flux [m^-2 s^-1]
    """
    SOL_momentum_loss_fraction = calc_SOL_momentum_loss_fraction(SOL_momentum_loss_function, target_electron_temp)

    upstream_electron_temp = calc_upstream_electron_temp(
        target_electron_temp=target_electron_temp,
        parallel_heat_flux_density=parallel_heat_flux_density,
        parallel_connection_length=parallel_connection_length,
        SOL_conduction_fraction=SOL_conduction_fraction,
        kappa_e0=kappa_e0,
    )

    upstream_total_pressure = calc_upstream_total_pressure(
        upstream_electron_density=upstream_electron_density,
        upstream_electron_temp=upstream_electron_temp,
        upstream_ratio_of_ion_to_electron_temp=upstream_ratio_of_ion_to_electron_temp,
        upstream_ratio_of_electron_to_ion_density=upstream_ratio_of_electron_to_ion_density,
        upstream_mach_number=upstream_mach_number,
    )

    f_basic_kwargs = dict(
        fuel_average_mass_number=fuel_average_mass_number,
        parallel_heat_flux_density=parallel_heat_flux_density,
        upstream_total_pressure=upstream_total_pressure,
        sheath_heat_transmission_factor=sheath_heat_transmission_factor,
    )

    f_other_kwargs = dict(
        target_ratio_of_ion_to_electron_temp=target_ratio_of_ion_to_electron_temp,
        target_ratio_of_electron_to_ion_density=target_ratio_of_electron_to_ion_density,
        target_mach_number=target_mach_number,
        toroidal_flux_expansion=toroidal_flux_expansion,
    )

    SOL_power_loss_fraction = calc_required_SOL_power_loss_fraction(
        target_electron_temp_basic=calc_target_electron_temp_basic(**f_basic_kwargs),
        f_other_target_electron_temp=calc_f_other_target_electron_temp(**f_other_kwargs),
        SOL_momentum_loss_fraction=SOL_momentum_loss_fraction,
        required_target_electron_temp=target_electron_temp,
    )

    f_vol_loss_kwargs = dict(SOL_power_loss_fraction=SOL_power_loss_fraction, SOL_momentum_loss_fraction=SOL_momentum_loss_fraction)

    target_electron_density = calc_target_electron_density(
        target_electron_density_basic=calc_target_electron_density_basic(**f_basic_kwargs),
        f_vol_loss_target_electron_density=calc_f_vol_loss_target_electron_density(**f_vol_loss_kwargs),
        f_other_target_electron_density=calc_f_other_target_electron_density(**f_other_kwargs),
    )

    target_electron_flux = calc_target_electron_flux(
        target_electron_flux_basic=calc_target_electron_flux_basic(**f_basic_kwargs),
        f_vol_loss_target_electron_flux=calc_f_vol_loss_target_electron_flux(**f_vol_loss_kwargs),
        f_other_target_electron_flux=calc_f_other_target_electron_flux(**f_other_kwargs),
    )

    return SOL_power_loss_fraction, upstream_electron_temp, target_electron_density, target_electron_flux



"""Open the atomic data files and return corresponding xr.Datasets and interpolators."""
from pathlib import Path
from typing import Optional, Union

import numpy as np
import xarray as xr
from scipy.interpolate import RectBivariateSpline, RegularGridInterpolator  # type: ignore[import]

from ..named_options import Impurity
from ..unit_handling import convert_units, magnitude, ureg


def read_atomic_data(directory: Optional[Path] = None) -> dict[Impurity, xr.DataArray]:
    """Read the atomic data files and return a dictionary mapping Impurity keys to xr.DataArrays of atomic data."""
    if directory is None:
        directory = Path(__file__).parent

    atomic_data_files = {
        Impurity.Helium: directory / "helium.nc",
        Impurity.Lithium: directory / "lithium.nc",
        Impurity.Beryllium: directory / "beryllium.nc",
        Impurity.Carbon: directory / "carbon.nc",
        Impurity.Nitrogen: directory / "nitrogen.nc",
        Impurity.Oxygen: directory / "oxygen.nc",
        Impurity.Neon: directory / "neon.nc",
        Impurity.Argon: directory / "argon.nc",
        Impurity.Krypton: directory / "krypton.nc",
        Impurity.Xenon: directory / "xenon.nc",
        Impurity.Tungsten: directory / "tungsten.nc",
    }

    atomic_data = {}

    for key, file in atomic_data_files.items():
        if not file.exists():
            raise FileNotFoundError(f"Could not find the atomic data file {file.absolute()} for species {key}")

        ds = xr.open_dataset(file).pint.quantify()

        # Convert the dimensions from linear to log values
        ds["dim_log_electron_temperature"] = xr.DataArray(
            np.log10(magnitude(convert_units(ds.electron_temperature, ureg.eV))), dims="dim_electron_temperature"
        )
        ds["dim_log_electron_density"] = xr.DataArray(
            np.log10(magnitude(convert_units(ds.electron_density, ureg.m**-3))), dims="dim_electron_density"
        )
        ds["dim_log_ne_tau"] = xr.DataArray(np.log10(magnitude(convert_units(ds.ne_tau, ureg.m**-3 * ureg.s))), dims="dim_ne_tau")

        ds = ds.swap_dims({"dim_electron_temperature": "dim_log_electron_temperature"})
        ds = ds.swap_dims({"dim_electron_density": "dim_log_electron_density"})
        ds = ds.swap_dims({"dim_ne_tau": "dim_log_ne_tau"})

        def build_interpolator(curve: xr.Dataset) -> Union[RectBivariateSpline, RegularGridInterpolator]:
            if curve.ndim == 2:
                return RectBivariateSpline(
                    x=curve.dim_log_electron_temperature,
                    y=curve.dim_log_electron_density,
                    z=np.log10(magnitude(curve.transpose("dim_log_electron_temperature", "dim_log_electron_density"))),  # type: ignore[arg-type]
                )
            elif curve.ndim == 3:
                return RegularGridInterpolator(
                    points=(curve.dim_log_electron_temperature, curve.dim_log_electron_density, curve.dim_log_ne_tau),
                    values=np.log10(magnitude(curve.transpose("dim_log_electron_temperature", "dim_log_electron_density", "dim_log_ne_tau"))),  # type: ignore[arg-type]
                    bounds_error=False,
                    method="linear",
                )
            else:
                raise NotImplementedError(f"Cannot build an interpolator for a curve with ndim={curve.ndim}")

        ds = ds.assign_attrs(
            coronal_Lz_interpolator=build_interpolator(ds.coronal_electron_emission_prefactor),
            coronal_mean_Z_interpolator=build_interpolator(ds.coronal_mean_charge_state),
            noncoronal_Lz_interpolator=build_interpolator(ds.noncoronal_electron_emission_prefactor),
            noncoronal_mean_Z_interpolator=build_interpolator(ds.noncoronal_mean_charge_state),
        )

        # Drop the linear coordinates
        atomic_data[key] = ds.drop_vars(("dim_ne_tau", "dim_electron_temperature", "dim_electron_density"))

    return atomic_data



"""Divertor loading and functions to calculate OMP pitch (for q_parallel calculation)."""
import numpy as np
from scipy import constants  # type: ignore[import]

from ..unit_handling import ureg, wraps_ufunc


@wraps_ufunc(
    return_units=dict(B_pol_omp=ureg.T),
    input_units=dict(plasma_current=ureg.A, minor_radius=ureg.m),
)
def calc_B_pol_omp(plasma_current: float, minor_radius: float) -> float:
    """Calculate the poloidal magnetic field at the outboard midplane.

    Args:
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        minor_radius: [m] :term:`glossary link<minor_radius>`

    Returns:
         B_pol_omp [T]
    """
    return float(constants.mu_0 * plasma_current / (2.0 * np.pi * minor_radius))


@wraps_ufunc(
    return_units=dict(B_tor_omp=ureg.T),
    input_units=dict(magnetic_field_on_axis=ureg.T, major_radius=ureg.m, minor_radius=ureg.m),
)
def calc_B_tor_omp(magnetic_field_on_axis: float, major_radius: float, minor_radius: float) -> float:
    """Calculate the toroidal magnetic field at the outboard midplane.

    Args:
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        major_radius: [m] :term:`glossary link<major_radius>`
        minor_radius: [m] :term:`glossary link<minor_radius>`

    Returns:
         B_tor_omp [T]
    """
    return magnetic_field_on_axis * (major_radius / (major_radius + minor_radius))



"""Calculate the SOL power loss fraction required to achieve a specified target electron temperature."""

from typing import Union

import numpy as np
import xarray as xr

from ...unit_handling import Quantity


def calc_required_SOL_power_loss_fraction(
    target_electron_temp_basic: Union[Quantity, xr.DataArray],
    f_other_target_electron_temp: Union[float, xr.DataArray],
    SOL_momentum_loss_fraction: Union[Quantity, xr.DataArray],
    required_target_electron_temp: Union[Quantity, xr.DataArray],
) -> Union[float, xr.DataArray]:
    """Calculate the SOL radiated power fraction required to reach a desired target electron temperature.

    This equation is equation 15 of :cite:`stangeby_2018`, rearranged for $f_{cooling}$.

    Args:
        target_electron_temp_basic: from target_electron_temp module [eV]
        f_other_target_electron_temp: from target_electron_temp module [~]
        SOL_momentum_loss_fraction: fraction of momentum lost in SOL [~]
        required_target_electron_temp: what target temperature do we want? [eV]

    Returns:
        SOL_power_loss_fraction [~]
    """
    required_SOL_power_loss_fraction = xr.DataArray(
        1.0
        - np.sqrt(
            required_target_electron_temp
            / target_electron_temp_basic
            * (1.0 - SOL_momentum_loss_fraction) ** 2
            / f_other_target_electron_temp
        )
    )

    return required_SOL_power_loss_fraction.clip(min=0.0)



"""Methods to run analyses configured via input files."""

from pathlib import Path
from typing import Any, Union

import numpy as np
import xarray as xr
import yaml

from .algorithms import get_algorithm
from .algorithms.algorithm_class import Algorithm, CompositeAlgorithm
from .helpers import convert_named_options
from .unit_handling import set_default_units


def read_case(case: Union[str, Path]) -> tuple[dict[str, Any], Union[CompositeAlgorithm, Algorithm], dict[str, Any]]:
    """Read a yaml file corresponding to a given case.

    case should be passed either as a complete filepath to an input.yaml file or to
    the parent folder of an input.yaml file.
    """
    if Path(case).exists():
        case = Path(case)
        if case.is_dir():
            input_file = case / "input.yaml"
        else:
            input_file = case
    else:
        raise FileNotFoundError(f"Could not find {case}.")

    with open(input_file) as file:
        repr_d: dict[str, Any] = yaml.load(file, Loader=yaml.FullLoader)

    algorithms = repr_d.pop("algorithms")
    algorithm_list = [get_algorithm(algorithm) for algorithm in algorithms]

    # why doesn't mypy deduce the below without hint?
    algorithm: Union[Algorithm, CompositeAlgorithm] = CompositeAlgorithm(algorithm_list) if len(algorithm_list) > 1 else algorithm_list[0]

    points = repr_d.pop("points")

    grid_values = repr_d.pop("grid")
    for key, grid_spec in grid_values.items():
        grid_spacing = grid_spec.get("spacing", "linear")

        if grid_spacing == "linear":
            grid_vals = np.linspace(grid_spec["min"], grid_spec["max"], num=grid_spec["num"])
        elif grid_spacing == "log":
            grid_vals = np.logspace(np.log10(grid_spec["min"]), np.log10(grid_spec["max"]), num=grid_spec["num"])
        else:
            raise NotImplementedError(f"No implementation for grid with {grid_spec['spacing']} spacing.")

        repr_d[key] = xr.DataArray(grid_vals, coords={f"dim_{key}": grid_vals})

    for key, val in repr_d.items():
        if isinstance(val, (list, tuple)):
            repr_d[key] = [convert_named_options(key=key, val=v) for v in val]
        else:
            repr_d[key] = convert_named_options(key=key, val=val)

    for key, val in repr_d.items():
        repr_d[key] = set_default_units(key=key, value=val)

    return repr_d, algorithm, points



"""Routines to find the coordinates of the minimum or maximum value of a field."""
from collections.abc import Sequence
from typing import Optional

import numpy as np
import xarray as xr
from xarray.core.coordinates import DataArrayCoordinates

from .unit_handling import Quantity


def find_coords_of_minimum(array: xr.DataArray, keep_dims: Sequence[str] = [], mask: Optional[xr.DataArray] = None) -> DataArrayCoordinates:
    """Find the coordinates the minimum value of array.

    These coordinates can be used to find the value of other arrays at the same point.

    For example
    >>> import xarray as xr
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from cfspopcon.operational_space import find_coords_of_minimum
    >>> x = xr.DataArray(np.linspace(0, 1, num=10), dims="x")
    >>> y = xr.DataArray(np.linspace(-1, 1, num=20), dims="y")
    >>> z = xr.DataArray(np.abs(x + y), coords=dict(x=x, y=y))
    >>> z.T.plot()
    >>> line = z.isel(find_coords_of_minimum(z, keep_dims="y"))
    >>> plt.scatter(line.x, line.y)
    """
    large = Quantity(np.inf, array.pint.units)

    along_dims = [dim for dim in array.dims if dim not in keep_dims]

    if mask is None:
        point_coords = array.fillna(large).argmin(dim=along_dims)
    else:
        point_coords = array.where(mask).fillna(large).argmin(dim=along_dims)

    return point_coords  # type:ignore[return-value]


def find_coords_of_maximum(array: xr.DataArray, keep_dims: Sequence[str] = [], mask: Optional[xr.DataArray] = None) -> DataArrayCoordinates:
    """Find the coordinates of the maximum value of array."""
    return find_coords_of_minimum(-array, keep_dims=keep_dims, mask=mask)


def build_mask_from_dict(dataset: xr.Dataset, plot_params: dict) -> xr.DataArray:
    """Build a mask field which hides inaccessible parts of the operational space."""
    mask = xr.DataArray(True)

    for mask_key, mask_range in plot_params.get("where", dict()).items():

        mask_field = dataset[mask_key]
        mask_min = Quantity(mask_range.get("min", -np.inf), mask_range.get("units", ""))
        mask_max = Quantity(mask_range.get("max", +np.inf), mask_range.get("units", ""))

        mask = mask & ((mask_field > mask_min) & (mask_field < mask_max))

    return mask



"""Switch to the LOC scaling if it predicts a worse energy confinement than our selected tau_e scaling."""
import xarray as xr

from .. import formulas, named_options
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "energy_confinement_time",
    "P_in",
    "SOC_LOC_ratio",
]


def run_use_LOC_tau_e_below_threshold(
    plasma_stored_energy: Unitfull,
    energy_confinement_time: Unitfull,
    P_in: Unitfull,
    average_electron_density: Unitfull,
    confinement_time_scalar: Unitfull,
    plasma_current: Unitfull,
    magnetic_field_on_axis: Unitfull,
    major_radius: Unitfull,
    areal_elongation: Unitfull,
    separatrix_elongation: Unitfull,
    inverse_aspect_ratio: Unitfull,
    fuel_average_mass_number: Unitfull,
    triangularity_psi95: Unitfull,
    separatrix_triangularity: Unitfull,
    q_star: Unitfull,
) -> dict[str, Unitfull]:
    """Switch to the LOC scaling if it predicts a worse energy confinement than our selected tau_e scaling.

    Args:
        plasma_stored_energy: :term:`glossary link<plasma_stored_energy>`
        energy_confinement_time: :term:`glossary link<energy_confinement_time>`
        P_in: :term:`glossary link<P_in>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        confinement_time_scalar: :term:`glossary link<confinement_time_scalar>`
        plasma_current: :term:`glossary link<plasma_current>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        major_radius: :term:`glossary link<major_radius>`
        areal_elongation: :term:`glossary link<areal_elongation>`
        separatrix_elongation: :term:`glossary link<separatrix_elongation>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        fuel_average_mass_number: :term:`glossary link<fuel_average_mass_number>`
        triangularity_psi95: :term:`glossary link<triangularity_psi95>`
        separatrix_triangularity: :term:`glossary link<separatrix_triangularity>`
        q_star: :term:`glossary link<q_star>`

    Returns:
    :term:`energy_confinement_time`, :term:`P_in`, :term:`SOC_LOC_ratio`

    """
    # Calculate linear ohmic confinement for low density
    energy_confinement_time_LOC, P_in_LOC = formulas.calc_tau_e_and_P_in_from_scaling(
        confinement_time_scalar=confinement_time_scalar,
        plasma_current=plasma_current,
        magnetic_field_on_axis=magnetic_field_on_axis,
        average_electron_density=average_electron_density,
        major_radius=major_radius,
        areal_elongation=areal_elongation,
        separatrix_elongation=separatrix_elongation,
        inverse_aspect_ratio=inverse_aspect_ratio,
        fuel_average_mass_number=fuel_average_mass_number,
        triangularity_psi95=triangularity_psi95,
        separatrix_triangularity=separatrix_triangularity,
        plasma_stored_energy=plasma_stored_energy,
        q_star=q_star,
        tau_e_scaling=named_options.ConfinementScaling.LOC,
    )

    # Use Linearized Ohmic Confinement if it gives worse energy confinement.
    SOC_LOC_ratio = energy_confinement_time / energy_confinement_time_LOC
    energy_confinement_time = xr.where(
        SOC_LOC_ratio > 1.0, energy_confinement_time_LOC, energy_confinement_time
    )  # type:ignore[no-untyped-call]
    P_in = xr.where(SOC_LOC_ratio > 1.0, P_in_LOC, P_in)  # type:ignore[no-untyped-call]

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


use_LOC_tau_e_below_threshold = Algorithm(
    function=run_use_LOC_tau_e_below_threshold,
    return_keys=RETURN_KEYS,
)



#!.venv/bin/python
# Run this script from the repository directory.
"""CLI for cfspopcon."""
import sys
from pathlib import Path

import click
import matplotlib.pyplot as plt  # type:ignore[import]
import xarray as xr
from ipdb import launch_ipdb_on_exception  # type:ignore[import]

from cfspopcon import file_io
from cfspopcon.input_file_handling import read_case
from cfspopcon.plotting import make_plot, read_plot_style


@click.command()
@click.argument("case", type=click.Path(exists=True))
@click.option(
    "--plots",
    "-p",
    type=click.Path(exists=True),
    multiple=True,
)
@click.option("--show", is_flag=True, help="Display an interactive figure of the result")
@click.option("--debug", is_flag=True, help="Enable the ipdb exception catcher")
def run_popcon_cli(case: str, plots: tuple[str], show: bool, debug: bool) -> None:
    """Run POPCON from the command line.

    This function uses "Click" to develop the command line interface. You can execute it using
    poetry run python cfspopcon/cli.py --help

    You can specify a set of plots to create by specifying a plot style file after `-p` on the command-line. Multiple entries are supported.
    """
    if show and not plots:
        print(f"Speficied show={show}, but did not specify a plot style, see --plots!")
        sys.exit(1)

    if not debug:
        run_popcon(case, plots, show)
    else:
        with launch_ipdb_on_exception():
            run_popcon(case, plots, show)


def run_popcon(case: str, plots: tuple[str], show: bool) -> None:
    """Run popcon case.

    Args:
        case: specify case to run (corresponding to a case in cases)
        plots: specify which plots to make (corresponding to a plot_style in plot_styles)
        show: show the resulting plots
    """
    input_parameters, algorithm, points = read_case(case)

    dataset = xr.Dataset(input_parameters)

    algorithm.validate_inputs(dataset)
    algorithm.update_dataset(dataset, in_place=True)

    output_dir = Path(case) / "output" if Path(case).is_dir() else Path(case).parent / "output"
    output_dir.mkdir(exist_ok=True)

    file_io.write_dataset_to_netcdf(dataset, filepath=output_dir / "dataset.nc")

    for point, point_params in points.items():
        file_io.write_point_to_file(dataset, point, point_params, output_dir=output_dir)

    # Plot the results
    for plot_style in plots:
        make_plot(dataset, read_plot_style(plot_style), points, title=input_parameters.get("plot_title", "POPCON"), output_dir=output_dir)

    print("Done")
    if show:
        plt.show()


if __name__ == "__main__":
    run_popcon_cli()



"""Functions for saving results to file and loading those files."""
import json
from pathlib import Path
from typing import Any

import xarray as xr

from .helpers import convert_named_options
from .point_selection import build_mask_from_dict, find_coords_of_minimum
from .unit_handling import convert_to_default_units, set_default_units


def sanitize_variable(val: xr.DataArray, key: str) -> xr.DataArray:
    """Strip units and Enum values from a variable so that it can be stored in a NetCDF file."""
    try:
        val = convert_to_default_units(val, key).pint.dequantify()
    except KeyError:
        pass

    if val.dtype == object:
        if val.size == 1:
            val = val.item().name
        else:
            val = xr.DataArray([v.name for v in val.values])

    return val


def write_dataset_to_netcdf(dataset: xr.Dataset, filepath: Path) -> None:
    """Write a dataset to a NetCDF file."""
    serialized_dataset = dataset.copy()
    for key in dataset.keys():
        assert isinstance(key, str)  # because hashable type of key is broader str but we  know it's str
        serialized_dataset[key] = sanitize_variable(dataset[key], key)

    for key in dataset.coords:
        assert isinstance(key, str)  # because hashable type of key is broader str but we  know it's str
        serialized_dataset[key] = sanitize_variable(dataset[key], key)

    serialized_dataset.to_netcdf(filepath)


def promote_variable(val: xr.DataArray, key: str) -> Any:
    """Add back in units and Enum values that were removed by sanitize_variable."""
    try:
        val = set_default_units(val, key)
    except KeyError:
        pass

    if val.dtype == object:
        if val.size == 1:
            return convert_named_options(key, val.item())
        else:
            return [convert_named_options(key, v) for v in val.values]

    return val


def read_dataset_from_netcdf(filepath: Path) -> xr.Dataset:
    """Open a dataset from a NetCDF file."""
    dataset = xr.open_dataset(filepath)

    for key in dataset.keys():
        assert isinstance(key, str)
        dataset[key] = promote_variable(dataset[key], key)

    for key in dataset.coords:
        if key == "dim_species":
            dataset[key] = promote_variable(dataset[key], key="impurity")
        else:
            assert isinstance(key, str)  # because hashable type of key is broader str but we  know it's str
            dataset[key] = promote_variable(dataset[key], key)

    return dataset


def write_point_to_file(dataset: xr.Dataset, point_key: str, point_params: dict, output_dir: Path) -> None:
    """Write the analysis values at the named points to a json file."""
    mask = build_mask_from_dict(dataset, point_params)

    if "minimize" not in point_params.keys() and "maximize" not in point_params.keys():
        raise ValueError(f"Need to provide either minimize or maximize in point specification. Keys were {point_params.keys()}")

    if "minimize" in point_params.keys():
        array = dataset[point_params["minimize"]]
    else:
        array = -dataset[point_params["maximize"]]

    point_coords = find_coords_of_minimum(array, keep_dims=point_params.get("keep_dims", []), mask=mask)

    point = dataset.isel(point_coords)

    for key in point.keys():
        assert isinstance(key, str)  # because hashable type of key is broader str but we  know it's str
        point[key] = sanitize_variable(point[key], key)

    for key in point.coords:
        assert isinstance(key, str)  # because hashable type of key is broader str but we  know it's str
        point[key] = sanitize_variable(point[key], key)

    output_dir.mkdir(parents=True, exist_ok=True)

    with open(output_dir / f"{point_key}.json", "w") as file:
        json.dump(point.to_dict(), file, indent=4)



"""Estimate 1D plasma profiles of density and temperature."""

from typing import Optional

import numpy as np
from numpy import float64
from numpy.typing import NDArray

from ..named_options import ProfileForm
from ..unit_handling import ureg, wraps_ufunc
from .plasma_profile_data.density_and_temperature_profile_fits import evaluate_density_and_temperature_profile_fits


@wraps_ufunc(
    return_units=dict(
        rho=ureg.dimensionless,
        electron_density_profile=ureg.n19,
        fuel_ion_density_profile=ureg.n19,
        electron_temp_profile=ureg.keV,
        ion_temp_profile=ureg.keV,
    ),
    input_units=dict(
        profile_form=None,
        average_electron_density=ureg.n19,
        average_electron_temp=ureg.keV,
        average_ion_temp=ureg.keV,
        electron_density_peaking=ureg.dimensionless,
        ion_density_peaking=ureg.dimensionless,
        temperature_peaking=ureg.dimensionless,
        dilution=ureg.dimensionless,
        normalized_inverse_temp_scale_length=ureg.dimensionless,
        npoints=None,
    ),
    output_core_dims=[("dim_rho",), ("dim_rho",), ("dim_rho",), ("dim_rho",), ("dim_rho",)],
)
def calc_1D_plasma_profiles(
    profile_form: ProfileForm,
    average_electron_density: float,
    average_electron_temp: float,
    average_ion_temp: float,
    electron_density_peaking: float,
    ion_density_peaking: float,
    temperature_peaking: float,
    dilution: float,
    normalized_inverse_temp_scale_length: Optional[float] = None,
    npoints: int = 50,
) -> tuple[NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64]]:
    """Estimate density and temperature profiles.

    Args:
        profile_form: select between analytic fit or profiles from Pablo Rodriguez-Fernandez
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        electron_density_peaking: [~] :term:`glossary link<electron_density_peaking>`
        ion_density_peaking: [~] :term:`glossary link<ion_density_peaking>`
        temperature_peaking: [~] :term:`glossary link<temperature_peaking>`
        dilution: dilution of main ions [~]
        normalized_inverse_temp_scale_length: [~] :term:`glossary link<normalized_inverse_temp_scale_length>`
        npoints: number of points to return in profile

    Returns:
        :term:`rho` [~], :term:`electron_density_profile` [1e19 m^-3], fuel_ion_density_profile [1e19 m^-3], :term:`electron_temp_profile` [keV], :term:`ion_temp_profile` [keV]

    Raises:
        ValueError: if profile_form == prf and normalized_inverse_temp_scale_length is not set
    """
    if profile_form == ProfileForm.analytic:
        rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile = calc_analytic_profiles(
            average_electron_density,
            average_electron_temp,
            average_ion_temp,
            electron_density_peaking,
            ion_density_peaking,
            temperature_peaking,
            dilution,
            npoints=npoints,
        )

    elif profile_form == ProfileForm.prf:
        if normalized_inverse_temp_scale_length is None:
            raise ValueError("normalized_inverse_temp_scale_length must be set if using profile_form = prf")

        rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile = calc_prf_profiles(
            average_electron_density,
            average_electron_temp,
            average_ion_temp,
            electron_density_peaking,
            ion_density_peaking,
            temperature_peaking,
            dilution,
            normalized_inverse_temp_scale_length,
            npoints=npoints,
        )

    return rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile


def calc_analytic_profiles(
    average_electron_density: float,
    average_electron_temp: float,
    average_ion_temp: float,
    electron_density_peaking: float,
    ion_density_peaking: float,
    temperature_peaking: float,
    dilution: float,
    npoints: int = 50,
) -> tuple[NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64]]:
    """Estimate density and temperature profiles using a simple analytic fit.

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        electron_density_peaking: [~] :term:`glossary link<electron_density_peaking>`
        ion_density_peaking: [~] :term:`glossary link<ion_density_peaking>`
        temperature_peaking: [~] :term:`glossary link<temperature_peaking>`
        dilution: dilution of main ions [~]
        npoints: number of points to return in profile

    Returns:
        :term:`rho` [~], :term:`electron_density_profile` [1e19 m^-3], fuel_ion_density_profile [1e19 m^-3], :term:`electron_temp_profile` [keV], :term:`ion_temp_profile` [keV]
    """
    rho = np.linspace(0, 1, num=npoints, endpoint=False)

    electron_density_profile = (
        average_electron_density * electron_density_peaking * ((1.0 - rho**2.0) ** (electron_density_peaking - 1.0))
    )
    fuel_ion_density_profile = (
        average_electron_density * dilution * (ion_density_peaking) * ((1.0 - rho**2.0) ** (ion_density_peaking - 1.0))
    )
    electron_temp_profile = average_electron_temp * temperature_peaking * ((1.0 - rho**2.0) ** (temperature_peaking - 1.0))
    ion_temp_profile = average_ion_temp * temperature_peaking * ((1.0 - rho**2.0) ** (temperature_peaking - 1.0))

    return rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile


def calc_prf_profiles(
    average_electron_density: float,
    average_electron_temp: float,
    average_ion_temp: float,
    electron_density_peaking: float,
    ion_density_peaking: float,
    temperature_peaking: float,
    dilution: float,
    normalized_inverse_temp_scale_length: float,
    npoints: int = 50,
) -> tuple[NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64]]:
    """Estimate density and temperature profiles using profiles from Pablo Rodriguez-Fernandez.

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        electron_density_peaking: [~] :term:`glossary link<electron_density_peaking>`
        ion_density_peaking: [~] :term:`glossary link<ion_density_peaking>`
        temperature_peaking: [~] :term:`glossary link<temperature_peaking>`
        dilution: dilution of main ions [~]
        normalized_inverse_temp_scale_length: [~] :term:`glossary link<normalized_inverse_temp_scale_length>`
        npoints: number of points to return in profile

    Returns:
        :term:`rho` [~], :term:`electron_density_profile` [1e19 m^-3], fuel_ion_density_profile [1e19 m^-3], :term:`electron_temp_profile` [keV], :term:`ion_temp_profile` [keV]
    """
    rho = np.linspace(0, 1, num=npoints, endpoint=False)

    rho, electron_temp_profile, electron_density_profile = evaluate_density_and_temperature_profile_fits(
        average_electron_temp,
        average_electron_density,
        temperature_peaking,
        electron_density_peaking,
        aLT=normalized_inverse_temp_scale_length,
        rho=rho,
        dataset="PRF",
    )
    rho, ion_temp_profile, fuel_ion_density_profile = evaluate_density_and_temperature_profile_fits(
        average_ion_temp,
        average_electron_density * dilution,
        temperature_peaking,
        ion_density_peaking,
        aLT=normalized_inverse_temp_scale_length,
        rho=rho,
        dataset="PRF",
    )

    return rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile



"""Calculate the fusion power and thermal gain (Q)."""
from .. import formulas, named_options
from ..unit_handling import Unitfull, convert_to_default_units, ureg
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "P_fusion",
    "P_neutron",
    "P_alpha",
    "P_external",
    "P_launched",
    "Q",
    "neutron_power_flux_to_walls",
    "neutron_rate",
]


def run_calc_fusion_gain(
    fusion_reaction: named_options.ReactionType,
    ion_temp_profile: Unitfull,
    heavier_fuel_species_fraction: Unitfull,
    fuel_ion_density_profile: Unitfull,
    rho: Unitfull,
    plasma_volume: Unitfull,
    surface_area: Unitfull,
    P_in: Unitfull,
    fraction_of_external_power_coupled: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate the fusion power and thermal gain (Q).

    Args:
        fusion_reaction: :term:`glossary link<fusion_reaction>`
        ion_temp_profile: :term:`glossary link<ion_temp_profile>`
        heavier_fuel_species_fraction: :term:`glossary link<heavier_fuel_species_fraction>`
        fuel_ion_density_profile: :term:`glossary link<fuel_ion_density_profile>`
        rho: :term:`glossary link<rho>`
        plasma_volume: :term:`glossary link<plasma_volume>`
        surface_area: :term:`glossary link<surface_area>`
        P_in: :term:`glossary link<P_in>`
        fraction_of_external_power_coupled: :term:`glossary link<fraction_of_external_power_coupled>`

    Returns:
        :term:`P_fusion`, :term:`P_neutron`, :term:`P_alpha`, :term:`P_external`, :term:`P_launched`, :term:`Q`, :term:`neutron_power_flux_to_walls` :term:`neutron_rate`
    """
    P_fusion, P_neutron, P_alpha = formulas.calc_fusion_power(
        fusion_reaction, ion_temp_profile, heavier_fuel_species_fraction, fuel_ion_density_profile, rho, plasma_volume
    )

    P_external = (P_in - P_alpha).clip(min=0.0 * ureg.MW)
    P_launched = P_external / fraction_of_external_power_coupled
    Q = formulas.thermal_calc_gain_factor(P_fusion, P_launched)

    neutron_power_flux_to_walls, neutron_rate = formulas.calc_neutron_flux_to_walls(
        P_neutron, surface_area, fusion_reaction, ion_temp_profile, heavier_fuel_species_fraction
    )

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_fusion_gain = Algorithm(
    function=run_calc_fusion_gain,
    return_keys=RETURN_KEYS,
)



"""Routines to calculate the combined electron and ion pressure in the SOL."""
from typing import Union

import xarray as xr

from ...unit_handling import Quantity


def calc_upstream_total_pressure(
    upstream_electron_density: Union[Quantity, xr.DataArray],
    upstream_electron_temp: Union[Quantity, xr.DataArray],
    upstream_ratio_of_ion_to_electron_temp: Union[float, xr.DataArray],
    upstream_ratio_of_electron_to_ion_density: Union[float, xr.DataArray],
    upstream_mach_number: Union[float, xr.DataArray] = 0.0,
) -> Union[Quantity, xr.DataArray]:
    """Calculate the upstream total pressure (including the ion temperature contribution).

    Same as calc_total_pressure, but with a default value upstream_mach_number=0.0.

    Args:
        upstream_electron_density: [m^-3]
        upstream_electron_temp: [eV]
        upstream_ratio_of_ion_to_electron_temp: tau_t = (T_i / T_e) [~]
        upstream_ratio_of_electron_to_ion_density: z_t = (ne / ni) [~]
        upstream_mach_number: M_t = (parallel ion velocity / sound speed) [~]

    Returns:
        upstream_total_pressure [atm]
    """
    return calc_total_pressure(
        electron_density=upstream_electron_density,
        electron_temp=upstream_electron_temp,
        ratio_of_ion_to_electron_temp=upstream_ratio_of_ion_to_electron_temp,
        ratio_of_electron_to_ion_density=upstream_ratio_of_electron_to_ion_density,
        mach_number=upstream_mach_number,
    )


def calc_total_pressure(
    electron_density: Union[Quantity, xr.DataArray],
    electron_temp: Union[Quantity, xr.DataArray],
    ratio_of_ion_to_electron_temp: Union[float, xr.DataArray],
    ratio_of_electron_to_ion_density: Union[float, xr.DataArray],
    mach_number: Union[float, xr.DataArray],
) -> Union[Quantity, xr.DataArray]:
    """Calculate the total pressure (including ion temperature contribution).

    From equation 20, :cite:`stangeby_2018`.

    Args:
        electron_density: [m^-3]
        electron_temp: [eV]
        ratio_of_ion_to_electron_temp: tau_t = (T_i / T_e) [~]
        ratio_of_electron_to_ion_density: z_t = (ne / ni) [~]
        mach_number: M_t = (parallel ion velocity / sound speed) [~]

    Returns:
        upstream_total_pressure [atm]
    """
    return (
        (1.0 + mach_number**2)
        * electron_density
        * electron_temp
        * (1.0 + ratio_of_ion_to_electron_temp / ratio_of_electron_to_ion_density)
    )



"""Calculate the concentration and effect of a core radiator required to achieve above a defined core radiative fraction."""
import numpy as np
import xarray as xr

from .. import formulas, named_options
from ..atomic_data import read_atomic_data
from ..helpers import make_impurities_array
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "core_radiator_concentration",
    "P_radiated_by_core_radiator",
    "P_radiation",
    "core_radiator_concentration",
    "core_radiator_charge_state",
    "zeff_change_from_core_rad",
    "dilution_change_from_core_rad",
    "z_effective",
    "dilution",
]


def run_calc_extrinsic_core_radiator(
    minimum_core_radiated_fraction: Unitfull,
    P_in: Unitfull,
    P_radiation: Unitfull,
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    z_effective: Unitfull,
    dilution: Unitfull,
    rho: Unitfull,
    electron_density_profile: Unitfull,
    electron_temp_profile: Unitfull,
    plasma_volume: Unitfull,
    radiated_power_method: named_options.RadiationMethod,
    radiated_power_scalar: Unitfull,
    core_radiator: named_options.Impurity,
) -> dict[str, Unitfull]:
    """Calculate the concentration and effect of a core radiator required to achieve above a defined core radiative fraction.

    Args:
        minimum_core_radiated_fraction: :term:`glossary link<minimum_core_radiated_fraction>`
        P_in: :term:`glossary link<P_in>`
        P_radiation: :term:`glossary link<P_radiation>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        average_electron_temp: :term:`glossary link<average_electron_temp>`
        z_effective: :term:`glossary link<z_effective>`
        dilution: :term:`dilution`
        rho: :term:`glossary link<rho>`
        electron_density_profile: :term:`glossary link<electron_density_profile>`
        electron_temp_profile: :term:`glossary link<electron_temp_profile>`
        plasma_volume: :term:`glossary link<plasma_volume>`
        radiated_power_method: :term:`glossary link<radiated_power_method>`
        radiated_power_scalar: :term:`radiated_power_scalar`
        core_radiator: :term:`glossary link<core_radiator>`

    Returns:
        :term:`core_radiator_concentration`, :term:`P_radiated_by_core_radiator`, :term:`P_radiation`, :term:`core_radiator_concentration`, :term:`core_radiator_charge_state`, :term:`zeff_change_from_core_rad` :term:`dilution_change_from_core_rad`, :term:`z_effective`, :term:`dilution`

    """
    atomic_data = read_atomic_data()

    # Force P_radiated_by_core_radiator to be >= 0.0 (core radiator cannot reduce radiated power)
    P_radiated_by_core_radiator = np.maximum(minimum_core_radiated_fraction * P_in - P_radiation, 0.0)
    P_radiation = np.maximum(minimum_core_radiated_fraction * P_in, P_radiation)

    P_rad_per_core_radiator = radiated_power_scalar * formulas.calc_impurity_radiated_power(
        radiated_power_method=named_options.RadiationMethod.Radas
        if radiated_power_method == named_options.RadiationMethod.Inherent
        else radiated_power_method,
        rho=rho,
        electron_temp_profile=electron_temp_profile,
        electron_density_profile=electron_density_profile,
        impurities=make_impurities_array(core_radiator, 1.0),
        plasma_volume=plasma_volume,
        atomic_data=atomic_data,
    ).sum(dim="dim_species")
    core_radiator_concentration = xr.where(  # type:ignore[no-untyped-call]
        P_radiated_by_core_radiator > 0, P_radiated_by_core_radiator / P_rad_per_core_radiator, 0.0
    )

    core_radiator_charge_state = formulas.calc_impurity_charge_state(
        average_electron_density, average_electron_temp, core_radiator, atomic_data
    )
    zeff_change_from_core_rad = formulas.calc_change_in_zeff(core_radiator_charge_state, core_radiator_concentration)
    dilution_change_from_core_rad = formulas.calc_change_in_dilution(core_radiator_charge_state, core_radiator_concentration)

    z_effective = z_effective + zeff_change_from_core_rad
    dilution = (dilution - dilution_change_from_core_rad).clip(min=0.0)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_extrinsic_core_radiator = Algorithm(
    function=run_calc_extrinsic_core_radiator,
    return_keys=RETURN_KEYS,
)



"""Handling of yaml plot configuration."""
from pathlib import Path
from typing import Any, Union

import yaml


def read_plot_style(plot_style: Union[str, Path]) -> dict[str, Any]:
    """Read a yaml file corresponding to a given plot_style.

    plot_style may be passed either as a complete filepath or as a string matching a plot_style in "plot_styles"
    """
    if Path(plot_style).exists():
        input_file = plot_style
    else:
        raise FileNotFoundError(f"Could not find {plot_style}!")

    with open(input_file) as file:
        repr_d: dict[str, Any] = yaml.load(file, Loader=yaml.FullLoader)

    return repr_d



"""Physics calculations & lumped-parameter models."""
from importlib.metadata import metadata

__version__ = metadata(__package__)["Version"]
__author__ = metadata(__package__)["Author"]

from . import algorithms, file_io, formulas, helpers, named_options, unit_handling
from .algorithms.algorithm_class import Algorithm, CompositeAlgorithm
from .input_file_handling import read_case
from .plotting import read_plot_style
from .point_selection import find_coords_of_maximum, find_coords_of_minimum
from .unit_handling import (
    Quantity,
    Unit,
    convert_to_default_units,
    convert_units,
    default_unit,
    magnitude,
    magnitude_in_default_units,
    set_default_units,
    ureg,
    wraps_ufunc,
)

# export main classes users should need as well as the option enums
__all__ = [
    "helpers",
    "named_options",
    "algorithms",
    "formulas",
    "unit_handling",
    "ureg",
    "Quantity",
    "Unit",
    "wraps_ufunc",
    "magnitude_in_default_units",
    "set_default_units",
    "default_unit",
    "convert_to_default_units",
    "convert_units",
    "magnitude",
    "read_case",
    "read_plot_style",
    "find_coords_of_maximum",
    "find_coords_of_minimum",
    "Algorithm",
    "CompositeAlgorithm",
    "file_io",
]



"""Estimate the density peaking based on scaling from C. Angioni."""
import numpy as np
import xarray as xr

from ..unit_handling import Unitfull, ureg, wraps_ufunc


def calc_density_peaking(effective_collisionality: Unitfull, betaE: Unitfull, nu_noffset: Unitfull) -> Unitfull:
    """Calculate the density peaking (peak over volume average).

    Equation 3 from p1334 of Angioni et al, "Scaling of density peaking in H-mode plasmas based on a combined
    database of AUG and JET observations" :cite:`angioni_scaling_2007`

    Args:
        effective_collisionality: [~] :term:`glossary link <effective_collisionality>`
        betaE: beta due to external field [~]
        nu_noffset: scalar offset added to peaking factor [~]

    Returns:
        :term:`nu_n` [~]
    """
    nu_n = (1.347 - 0.117 * np.log(effective_collisionality) - 4.03 * betaE) + nu_noffset
    if isinstance(nu_n, xr.DataArray):
        return nu_n.clip(1.0, float("inf"))
    else:
        return max(nu_n, 1.0 * ureg.dimensionless)


@wraps_ufunc(
    return_units=dict(effective_collisionality=ureg.dimensionless),
    input_units=dict(
        average_electron_density=ureg.n19, average_electron_temp=ureg.keV, major_radius=ureg.m, z_effective=ureg.dimensionless
    ),
)
def calc_effective_collisionality(
    average_electron_density: float, average_electron_temp: float, major_radius: float, z_effective: float
) -> float:
    """Calculate the effective collisionality.

    From p1327 of Angioni et al, "Scaling of density peaking in H-mode plasmas based on a combined
    database of AUG and JET observations" :cite:`angioni_scaling_2007`

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        major_radius: [m] :term:`glossary link<major_radius>`
        z_effective: [~] :term:`glossary link<z_effective>`

    Returns:
        :term:`effective_collisionality` [~]
    """
    return float((0.1 * z_effective * average_electron_density * major_radius) / (average_electron_temp**2.0))



"""Calculate the ratio of magnetic to plasma (kinetic) pressure."""
import numpy as np

from ..unit_handling import Quantity, Unitfull, convert_units, ureg


def _calc_beta_general(
    average_electron_density: Unitfull, average_electron_temp: Unitfull, average_ion_temp: Unitfull, magnetic_field: Unitfull
) -> Unitfull:
    """Calculate the average ratio of the plasma pressure to the magnetic pressure due to a magnetic_field.

    Using equation 11.58 from Freidberg, "Plasma Physics and Fusion Energy" :cite:`freidberg_plasma_2007`

    The unit_conversion_factor comes from cancelling the units to get a dimensionless quantity

        >>> from pint import Quantity
        >>> n = Quantity(1e19, "m^-3")
        >>> T = Quantity(1, "keV")
        >>> B = Quantity(1, "T")
        >>> mu_0 = Quantity(1, "mu_0")
        >>> ((2*mu_0 * n * T / (B**2)).to('').units
        <Unit('dimensionless')>

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        magnetic_field: magnetic field generating magnetic pressure [T]

    Returns:
         beta (toroidal or poloidal) [~]
    """
    mu_0 = Quantity(1, "mu_0")
    # to make the result dimensionless
    unit_conversion_factor = 2 * mu_0
    ret = unit_conversion_factor * (average_electron_density * (average_electron_temp + average_ion_temp)) / (magnetic_field**2)
    return convert_units(ret, ureg.dimensionless)


def calc_beta_toroidal(
    average_electron_density: Unitfull, average_electron_temp: Unitfull, average_ion_temp: Unitfull, magnetic_field_on_axis: Unitfull
) -> Unitfull:
    """Calculate the average ratio of the plasma pressure to the magnetic pressure due to the toroidal field.

    Also called beta_external, since the toroidal field is generated by external toroidal field coils.
    Using equation 11.58 from Freidberg, "Plasma Physics and Fusion Energy" :cite:`freidberg_plasma_2007`

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`

    Returns:
         :term:`beta_toroidal` [~]
    """
    return _calc_beta_general(average_electron_density, average_electron_temp, average_ion_temp, magnetic_field=magnetic_field_on_axis)


def calc_beta_poloidal(
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    average_ion_temp: Unitfull,
    plasma_current: Unitfull,
    minor_radius: Unitfull,
) -> Unitfull:
    """Calculate the average ratio of the plasma pressure to the magnetic pressure due to the plasma current.

    Calculates the poloidal magnetic field at radius a from the plasma current using
    equation 11.55 from Freidberg, "Plasma Physics and Fusion Energy" :cite:`freidberg_plasma_2007`
    and then evaluates beta_poloidal using
    equation 11.58 from Freidberg, "Plasma Physics and Fusion Energy" :cite:`freidberg_plasma_2007`

    The unit_conversion_factor cancels the units, and can be calculated using the following

        >>> from pint import Quantity
        >>> from numpy import pi
        >>> mu_0 = Quantity(1, "mu_0")
        >>> I = Quantity(1, "MA")
        >>> minor_radius = Quantity(1, "m")
        >>> (mu_0 * I / (2 * pi * minor_radius)).to("T").units
        <Unit('tesla')>

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        average_ion_temp: [keV] :term:`glossary link<average_ion_temp>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        minor_radius: [m] :term:`glossary link<minor_radius>`

    Returns:
        :term:`beta_poloidal` [~]
    """
    mu_0 = Quantity(1, "mu_0")
    # to ensure the final result is in units of tesla
    units_conversion_factor = mu_0 / (2 * np.pi)
    B_pol = units_conversion_factor * plasma_current / minor_radius

    return _calc_beta_general(average_electron_density, average_electron_temp, average_ion_temp, magnetic_field=B_pol)


def calc_beta_total(beta_toroidal: Unitfull, beta_poloidal: Unitfull) -> Unitfull:
    """Calculate the total beta from the toroidal and poloidal betas.

    Using equation 11.59 from Freidberg, "Plasma Physics and Fusion Energy" :cite:`freidberg_plasma_2007`

    Args:
        beta_toroidal: [~] :term:`glossary link<beta_toroidal>`
        beta_poloidal: [~] :term:`glossary link<beta_poloidal>`

    Returns:
         :term:`beta_total` [~]
    """
    return 1.0 / (1.0 / beta_toroidal + 1.0 / beta_poloidal)


def calc_beta_normalised(beta: Unitfull, minor_radius: Unitfull, magnetic_field_on_axis: Unitfull, plasma_current: Unitfull) -> Unitfull:
    """Normalize beta to stability (Troyon) parameters.

    See section 6.18 in Wesson :cite:`wesson_tokamaks_2011`.

    Args:
        beta: plasma pressure normalized against toroidal B-on-axis [%]
        minor_radius: [m] :term:`glossary link<minor_radius>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        plasma_current: [MA] :term:`glossary link<plasma_current>`

    Returns:
         :term:`beta_N`
    """
    normalisation = plasma_current / (minor_radius * magnetic_field_on_axis)

    beta_N = beta / normalisation

    return beta_N



"""Calculate the inherent (Bremsstrahlung and Synchrotron) radiated power."""

import numpy as np
from numpy import float64
from numpy.typing import NDArray

from ...unit_handling import ureg, wraps_ufunc
from ..helpers import integrate_profile_over_volume


@wraps_ufunc(
    return_units=dict(P_rad_bremsstrahlung=ureg.MW),
    input_units=dict(
        rho=ureg.dimensionless,
        electron_density_profile=ureg.n19,
        electron_temp_profile=ureg.keV,
        z_effective=ureg.dimensionless,
        plasma_volume=ureg.m**3,
    ),
    input_core_dims=[("dim_rho",), ("dim_rho",), ("dim_rho",), (), ()],
)
def calc_bremsstrahlung_radiation(
    rho: NDArray[float64],
    electron_density_profile: NDArray[float64],
    electron_temp_profile: NDArray[float64],
    z_effective: float,
    plasma_volume: float,
) -> float:
    """Calculate the Bremsstrahlung radiated power due to the main plasma.

    Formula 13 in :cite:`stott_feasibility_2005`

    Args:
        electron_density_profile: [1e19 m^-3] :term:`glossary link<electron_density_profile>`
        electron_temp_profile: [keV] :term:`glossary link<electron_temp_profile>`
        z_effective: [~] :term:`glossary link<z_effective>`
        rho: [~] :term:`glossary link<rho>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
        Radiated bremsstrahlung power per cubic meter [MW / m^3]
    """
    ne20 = electron_density_profile / 10

    Tm = 511.0  # keV, Tm = m_e * c**2
    xrel = (1.0 + 2.0 * electron_temp_profile / Tm) * (
        1.0 + (2.0 / z_effective) * (1.0 - 1.0 / (1.0 + electron_temp_profile / Tm))
    )  # relativistic correction factor

    fKb = ne20**2 * np.sqrt(electron_temp_profile) * xrel
    Kb = integrate_profile_over_volume(fKb, rho, plasma_volume)  # radial profile factor

    P_brem = 5.35e-3 * z_effective * Kb  # volume-averaged bremsstrahlung radiaton in MW

    return P_brem


@wraps_ufunc(
    return_units=dict(P_rad_synchrotron=ureg.MW),
    input_units=dict(
        rho=ureg.dimensionless,
        electron_density_profile=ureg.n19,
        electron_temp_profile=ureg.keV,
        major_radius=ureg.m,
        minor_radius=ureg.m,
        magnetic_field_on_axis=ureg.T,
        separatrix_elongation=ureg.dimensionless,
        plasma_volume=ureg.m**3,
    ),
    input_core_dims=[("dim_rho",), ("dim_rho",), ("dim_rho",), (), (), (), (), ()],
)
def calc_synchrotron_radiation(
    rho: NDArray[float64],
    electron_density_profile: NDArray[float64],
    electron_temp_profile: NDArray[float64],
    major_radius: float,
    minor_radius: float,
    magnetic_field_on_axis: float,
    separatrix_elongation: float,
    plasma_volume: float,
) -> float:
    """Calculate the Synchrotron radiated power due to the main plasma.

    This can be an important loss mechanism in high temperature plasmas.

    Formula 15 in :cite:`stott_feasibility_2005`

    For now this calculation assumes 90% wall reflectivity, consistent with stott_feasibility_2005.

    This calculation also assumes profiles of the form n(r) = n[1 - (r/a)**2]**alpha_n and
    T(r) = Tedge + (T - Tedge)[1 - (r/a)**gamma_T]**alpha_T. For now, these are assumed as
    gamma_T = 2, alpha_n = 0.5 and alpha_T = 1, consistent with stott_feasibility_2005.

    An alternative approach could be developed using formula 6 in :cite:`zohm_use_2019`, which assumes 80% wall reflectivity.

    Args:
        electron_density_profile: [1e19 m^-3] :term:`glossary link<electron_density_profile>`
        electron_temp_profile: [keV] :term:`glossary link<electron_temp_profile>`
        major_radius: [m] :term:`glossary link<major_radius>`
        minor_radius: [m] :term:`glossary link<minor_radius>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        separatrix_elongation: [~] :term:`glossary link<separatrix_elongation>`
        rho: [~] :term:`glossary link<rho>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
        Radiated bremsstrahlung power per cubic meter [MW / m^3]
    """
    ne20 = electron_density_profile / 10

    Rw = 0.8  # wall reflectivity
    gamma_T = 2  # temperature profile inner exponent (2 is ~parabolic)
    alpha_n = 0.5  # density profile outer exponent (0.5 is rather broad)
    alpha_T = 1  # temperature profile outer exponent (1 is ~parabolic)

    # effective optical thickness
    rhoa = 6.04e3 * minor_radius * ne20 / magnetic_field_on_axis
    # profile peaking correction
    Ks = (
        (alpha_n + 3.87 * alpha_T + 1.46) ** (-0.79)
        * (1.98 + alpha_n) ** (1.36)
        * gamma_T**2.14
        * (gamma_T**1.53 + 1.87 * alpha_T - 0.16) ** (-1.33)
    )
    # aspect ratio correction
    Gs = 0.93 * (1 + 0.85 * np.exp(-0.82 * major_radius / minor_radius))

    # dimensionless parameter to account for plasma transparency and wall reflections
    Phi = (
        6.86e-5
        * separatrix_elongation ** (-0.21)
        * (16 + electron_temp_profile) ** (2.61)
        * ((rhoa / (1 - Rw)) ** (0.41) + 0.12 * electron_temp_profile) ** (-1.51)
        * Ks
        * Gs
    )

    P_sync_r = 6.25e-3 * ne20 * electron_temp_profile * magnetic_field_on_axis**2 * Phi
    P_sync = integrate_profile_over_volume(P_sync_r, rho, plasma_volume)

    return P_sync



"""Plotting functionality."""
from .coordinate_formatter import CoordinateFormatter
from .make_plot import label_contour, make_plot, units_to_string
from .plot_style_handling import read_plot_style

__all__ = [
    "CoordinateFormatter",
    "label_contour",
    "make_plot",
    "units_to_string",
    "read_plot_style",
]



"""Calculate the input power required to maintain the stored energy, given a tau_e scaling."""
from .. import formulas, named_options
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "energy_confinement_time",
    "P_in",
]


def run_calc_power_balance_from_tau_e(
    plasma_stored_energy: Unitfull,
    average_electron_density: Unitfull,
    confinement_time_scalar: Unitfull,
    plasma_current: Unitfull,
    magnetic_field_on_axis: Unitfull,
    major_radius: Unitfull,
    areal_elongation: Unitfull,
    separatrix_elongation: Unitfull,
    inverse_aspect_ratio: Unitfull,
    fuel_average_mass_number: Unitfull,
    triangularity_psi95: Unitfull,
    separatrix_triangularity: Unitfull,
    q_star: Unitfull,
    energy_confinement_scaling: named_options.ConfinementScaling,
) -> dict[str, Unitfull]:
    """Calculate the input power required to maintain the stored energy, given a tau_e scaling.

    Args:
        plasma_stored_energy: :term:`glossary link<plasma_stored_energy>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        confinement_time_scalar: :term:`glossary link<confinement_time_scalar>`
        plasma_current: :term:`glossary link<plasma_current>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        major_radius: :term:`glossary link<major_radius>`
        areal_elongation: :term:`glossary link<areal_elongation>`
        separatrix_elongation: :term:`glossary link<separatrix_elongation>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        fuel_average_mass_number: :term:`glossary link<fuel_average_mass_number>`
        triangularity_psi95: :term:`glossary link<triangularity_psi95>`
        separatrix_triangularity: :term:`glossary link<separatrix_triangularity>`
        q_star: :term:`glossary link<q_star>`
        energy_confinement_scaling: :term:`glossary link<energy_confinement_scaling>`

    Returns:
    :term:`energy_confinement_time`, :term:`P_in`

    """
    energy_confinement_time, P_in = formulas.calc_tau_e_and_P_in_from_scaling(
        confinement_time_scalar=confinement_time_scalar,
        plasma_current=plasma_current,
        magnetic_field_on_axis=magnetic_field_on_axis,
        average_electron_density=average_electron_density,
        major_radius=major_radius,
        areal_elongation=areal_elongation,
        separatrix_elongation=separatrix_elongation,
        inverse_aspect_ratio=inverse_aspect_ratio,
        fuel_average_mass_number=fuel_average_mass_number,
        triangularity_psi95=triangularity_psi95,
        separatrix_triangularity=separatrix_triangularity,
        plasma_stored_energy=plasma_stored_energy,
        q_star=q_star,
        tau_e_scaling=energy_confinement_scaling,
    )

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_power_balance_from_tau_e = Algorithm(
    function=run_calc_power_balance_from_tau_e,
    return_keys=RETURN_KEYS,
)



"""Compute all terms in the two-point-model for a fixed SOL power loss fraction."""
from typing import Union

import numpy as np
import xarray as xr

from ...named_options import MomentumLossFunction
from ...unit_handling import Quantity, Unitfull, ureg
from .momentum_loss_functions import calc_SOL_momentum_loss_fraction
from .target_electron_density import (
    calc_f_other_target_electron_density,
    calc_f_vol_loss_target_electron_density,
    calc_target_electron_density,
    calc_target_electron_density_basic,
)
from .target_electron_flux import (
    calc_f_other_target_electron_flux,
    calc_f_vol_loss_target_electron_flux,
    calc_target_electron_flux,
    calc_target_electron_flux_basic,
)
from .target_electron_temp import (
    calc_f_other_target_electron_temp,
    calc_f_vol_loss_target_electron_temp,
    calc_target_electron_temp,
    calc_target_electron_temp_basic,
)
from .total_pressure import calc_upstream_total_pressure
from .upstream_electron_temp import calc_upstream_electron_temp


def solve_two_point_model(
    SOL_power_loss_fraction: Unitfull,
    parallel_heat_flux_density: Unitfull,
    parallel_connection_length: Unitfull,
    upstream_electron_density: Unitfull,
    toroidal_flux_expansion: Unitfull,
    fuel_average_mass_number: Unitfull,
    kappa_e0: Unitfull,
    SOL_momentum_loss_function: Union[MomentumLossFunction, xr.DataArray],
    initial_target_electron_temp: Unitfull = 10.0 * ureg.eV,
    sheath_heat_transmission_factor: Unitfull = 7.5 * ureg.dimensionless,
    SOL_conduction_fraction: Unitfull = 1.0 * ureg.dimensionless,
    target_ratio_of_ion_to_electron_temp: Unitfull = 1.0 * ureg.dimensionless,
    target_ratio_of_electron_to_ion_density: Unitfull = 1.0 * ureg.dimensionless,
    target_mach_number: Unitfull = 1.0 * ureg.dimensionless,
    upstream_ratio_of_ion_to_electron_temp: Unitfull = 1.0 * ureg.dimensionless,
    upstream_ratio_of_electron_to_ion_density: Unitfull = 1.0 * ureg.dimensionless,
    upstream_mach_number: Unitfull = 0.0 * ureg.dimensionless,
    # Controlling the iterative solve
    max_iterations: int = 100,
    upstream_temp_relaxation: float = 0.5,
    target_electron_density_relaxation: float = 0.5,
    target_temp_relaxation: float = 0.5,
    upstream_temp_max_residual: float = 1e-2,
    target_electron_density_max_residual: float = 1e-2,
    target_temp_max_residual: float = 1e-2,
    # Return converged values even if not whole array is converged
    raise_error_if_not_converged: bool = True,
    # Print information about the solve to terminal
    quiet: bool = True,
) -> tuple[Union[Quantity, xr.DataArray], Union[Quantity, xr.DataArray], Union[Quantity, xr.DataArray], Union[Quantity, xr.DataArray]]:
    """Calculate the upstream and target electron temperature and target electron density according to the extended two-point-model.

    Args:
        SOL_power_loss_fraction: [~]
        parallel_heat_flux_density: [GW/m^2]
        parallel_connection_length: [m]
        upstream_electron_density: [m^-3]
        toroidal_flux_expansion: [~]
        fuel_average_mass_number: [~]
        kappa_e0: electron heat conductivity constant [W / (eV^3.5 * m)]
        SOL_momentum_loss_function: which momentum loss function to use
        initial_target_electron_temp: starting guess for target electron temp [eV]
        sheath_heat_transmission_factor: [~]
        SOL_conduction_fraction: [~]
        target_ratio_of_ion_to_electron_temp: [~]
        target_ratio_of_electron_to_ion_density: [~]
        target_mach_number: [~]
        upstream_ratio_of_ion_to_electron_temp: [~]
        upstream_ratio_of_electron_to_ion_density: [~]
        upstream_mach_number: [~]
        max_iterations: how many iterations to try before returning NaN
        upstream_temp_relaxation: step-size for upstream Te evolution
        target_electron_density_relaxation: step-size for target ne evolution
        target_temp_relaxation: step-size for target Te evolution
        upstream_temp_max_residual: relative rate of change for convergence for upstream Te evolution
        target_electron_density_max_residual: relative rate of change for convergence for target ne evolution
        target_temp_max_residual: relative rate of change for convergence for target Te evolution
        raise_error_if_not_converged: raise an error if not all point converge within max iterations (otherwise return NaN)
        quiet: if not True, print additional information about the iterative solve to terminal
    Returns:
        upstream_electron_temp [eV], target_electron_density [m^-3], target_electron_temp [eV], target_electron_flux [m^-2 s^-1]
    """
    f_other_kwargs = dict(
        target_ratio_of_ion_to_electron_temp=target_ratio_of_ion_to_electron_temp,
        target_ratio_of_electron_to_ion_density=target_ratio_of_electron_to_ion_density,
        target_mach_number=target_mach_number,
        toroidal_flux_expansion=toroidal_flux_expansion,
    )
    f_other_target_electron_density = calc_f_other_target_electron_density(**f_other_kwargs)
    f_other_target_electron_temp = calc_f_other_target_electron_temp(**f_other_kwargs)
    f_other_target_electron_flux = calc_f_other_target_electron_flux(**f_other_kwargs)

    iteration = 0
    target_electron_temp = initial_target_electron_temp

    while iteration < max_iterations:
        iteration += 1

        new_upstream_electron_temp = calc_upstream_electron_temp(
            target_electron_temp=target_electron_temp,
            parallel_heat_flux_density=parallel_heat_flux_density,
            parallel_connection_length=parallel_connection_length,
            SOL_conduction_fraction=SOL_conduction_fraction,
            kappa_e0=kappa_e0,
        )

        upstream_total_pressure = calc_upstream_total_pressure(
            upstream_electron_density=upstream_electron_density,
            upstream_electron_temp=new_upstream_electron_temp,
            upstream_ratio_of_ion_to_electron_temp=upstream_ratio_of_ion_to_electron_temp,
            upstream_ratio_of_electron_to_ion_density=upstream_ratio_of_electron_to_ion_density,
            upstream_mach_number=upstream_mach_number,
        )

        f_vol_loss_kwargs = dict(
            SOL_power_loss_fraction=SOL_power_loss_fraction,
            SOL_momentum_loss_fraction=calc_SOL_momentum_loss_fraction(SOL_momentum_loss_function, target_electron_temp),
        )

        f_basic_kwargs = dict(
            fuel_average_mass_number=fuel_average_mass_number,
            parallel_heat_flux_density=parallel_heat_flux_density,
            upstream_total_pressure=upstream_total_pressure,
            sheath_heat_transmission_factor=sheath_heat_transmission_factor,
        )

        target_electron_density_basic = calc_target_electron_density_basic(**f_basic_kwargs)
        target_electron_temp_basic = calc_target_electron_temp_basic(**f_basic_kwargs)

        f_vol_loss_target_electron_density = calc_f_vol_loss_target_electron_density(**f_vol_loss_kwargs)
        f_vol_loss_target_electron_temp = calc_f_vol_loss_target_electron_temp(**f_vol_loss_kwargs)

        new_target_electron_density = calc_target_electron_density(
            target_electron_density_basic=target_electron_density_basic,
            f_vol_loss_target_electron_density=f_vol_loss_target_electron_density,
            f_other_target_electron_density=f_other_target_electron_density,
        )

        new_target_electron_temp = calc_target_electron_temp(
            target_electron_temp_basic=target_electron_temp_basic,
            f_vol_loss_target_electron_temp=f_vol_loss_target_electron_temp,
            f_other_target_electron_temp=f_other_target_electron_temp,
        )

        if iteration == 1:
            upstream_electron_temp = new_upstream_electron_temp
            target_electron_density = new_target_electron_density
            target_electron_temp = new_target_electron_temp
            continue

        change_in_upstream_electron_temp = new_upstream_electron_temp - upstream_electron_temp
        change_in_target_electron_density = new_target_electron_density - target_electron_density
        change_in_target_electron_temp = new_target_electron_temp - target_electron_temp

        upstream_electron_temp = upstream_electron_temp + upstream_temp_relaxation * change_in_upstream_electron_temp
        target_electron_density = target_electron_density + target_electron_density_relaxation * change_in_target_electron_density
        target_electron_temp = target_electron_temp + target_temp_relaxation * change_in_target_electron_temp

        if np.all(
            [
                np.abs(change_in_upstream_electron_temp / upstream_electron_temp).max() < upstream_temp_max_residual,
                np.abs(change_in_target_electron_density / target_electron_density).max() < target_electron_density_max_residual,
                np.abs(change_in_target_electron_temp / target_electron_temp).max() < target_temp_max_residual,
            ]
        ):
            if not quiet:
                print(f"Converged in {iteration} iterations")
            break
    else:
        if raise_error_if_not_converged:
            raise RuntimeError("Iterative solve did not converge.")

    target_electron_flux_basic = calc_target_electron_flux_basic(**f_basic_kwargs)
    f_vol_loss_target_electron_flux = calc_f_vol_loss_target_electron_flux(**f_vol_loss_kwargs)

    target_electron_flux = calc_target_electron_flux(
        target_electron_flux_basic=target_electron_flux_basic,
        f_vol_loss_target_electron_flux=f_vol_loss_target_electron_flux,
        f_other_target_electron_flux=f_other_target_electron_flux,
    )

    mask = (
        (np.abs(change_in_upstream_electron_temp / upstream_electron_temp) < upstream_temp_max_residual)
        & (np.abs(change_in_target_electron_density / target_electron_density) < target_electron_density_max_residual)
        & (np.abs(change_in_target_electron_temp / target_electron_temp) < target_temp_max_residual)
    )

    number_nonconverged = np.count_nonzero(~mask)
    if number_nonconverged > 0 and not quiet:
        print(f"{number_nonconverged} values did not converge in {max_iterations} iterations.")

    upstream_electron_temp = xr.where(mask, upstream_electron_temp, np.nan)  # type:ignore[no-untyped-call]
    target_electron_density = xr.where(mask, target_electron_density, np.nan)  # type:ignore[no-untyped-call]
    target_electron_temp = xr.where(mask, target_electron_temp, np.nan)  # type:ignore[no-untyped-call]
    target_electron_flux = xr.where(mask, target_electron_flux, np.nan)  # type:ignore[no-untyped-call]

    return upstream_electron_temp, target_electron_density, target_electron_temp, target_electron_flux



"""Reaction energies and power densities."""

from typing import Any

from numpy import float64
from numpy.typing import NDArray
from scipy import constants  # type: ignore[import]


def reaction_energy_DT(
    sigmav: NDArray[float64], heavier_fuel_species_fraction: float
) -> tuple[float, float, float, NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Deuterium-Tritium reaction.

    Calculate reaction energies and power density values.

    Args:
        sigmav: :math:`\langle \sigma v \rangle` product in cm^3/s.
        heavier_fuel_species_fraction: n_Tritium / (n_Tritium + n_Deuterium) number fraction.

    Returns:
        Tuple of reaction energies and corresponding power densities.
    """
    rxn_energy: float = 17.6 * constants.value("electron volt")  # [MJ]
    rxn_energy_neut: float = rxn_energy * (4.0 / 5.0)  # [MJ]
    rxn_energy_charged: float = rxn_energy * (1.0 / 5.0)  # [MJ]
    convert_volume: float = 1e-6  # [m^3/cm^3]
    number_power_dens: NDArray[float64] = (
        sigmav * rxn_energy * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_neut: NDArray[float64] = (
        sigmav * rxn_energy_neut * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_charged: NDArray[float64] = (
        sigmav * rxn_energy_charged * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )

    # Units: [MJ], [MW*m^3]
    return rxn_energy, rxn_energy_neut, rxn_energy_charged, number_power_dens, number_power_dens_neut, number_power_dens_charged


def reaction_energy_DD(
    sigmav: tuple[
        NDArray[float64],
        NDArray[float64],
        NDArray[float64],
    ],
    **_: Any,
) -> tuple[NDArray[float64], float, NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Deuterium-Deuterium reaction.

    Calculate reaction energies and power density values.

    Args:
        sigmav: :math:`\langle \sigma v \rangle` product in cm^3/s.
        _: Unused placeholder to enable unified call syntax with e.g. :func:`reaction_energy_DT`.

    Returns:
        Tuple of reaction energies and corresponding power densities.
    """
    sigmav_tot, sigmav_1, sigmav_2 = sigmav
    path_1_energy: float = (1.01 + 3.02) * constants.value("electron volt")  # MJ, D+D -> p+T
    path_2_energy: float = (0.82 + 2.45) * constants.value("electron volt")  # MJ, D+D -> n+He3
    rxn_energy: NDArray[float64] = (path_1_energy * sigmav_1 + path_2_energy * sigmav_2) / sigmav_tot
    rxn_energy_neut: float = path_2_energy * (3.0 / 4.0)
    rxn_energy_charged: NDArray[float64] = (path_1_energy * sigmav_1 + (path_2_energy * (1.0 / 4.0)) * sigmav_2) / sigmav_tot

    # So number_power_dens*electron_density_profile**2 = power_dens [MW/m^3] no need to divide since nD=ne
    convert_volume = 1e-6  # m^3/cm^3
    number_power_dens: NDArray[float64] = (sigmav_1 * path_1_energy + sigmav_2 * path_2_energy) * convert_volume
    number_power_dens_neut: NDArray[float64] = sigmav_2 * rxn_energy_neut * convert_volume
    number_power_dens_charged: NDArray[float64] = (sigmav_1 * path_1_energy + sigmav_2 * path_2_energy * (1.0 / 4.0)) * convert_volume

    # Units: [MJ], [MW/m^3]
    return rxn_energy, rxn_energy_neut, rxn_energy_charged, number_power_dens, number_power_dens_neut, number_power_dens_charged


def reaction_energy_DHe3(
    sigmav: NDArray[float64], heavier_fuel_species_fraction: float
) -> tuple[float, float, float, NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Deuterium-Helium 3 reaction.

    Calculate reaction energies and power density values.

    Args:
        sigmav: :math:`\langle \sigma v \rangle` product in cm^3/s.
        heavier_fuel_species_fraction: n_heavier / (n_heavier + n_lighter) number fraction.

    Returns:
        Tuple of reaction energies and corresponding power densities.
    """
    rxn_energy: float = 18.3 * constants.value("electron volt")  # MJoules
    rxn_energy_neut: float = 0.0
    rxn_energy_charged: float = rxn_energy
    convert_volume: float = 1e-6  # m^3/cm^3
    number_power_dens: NDArray[float64] = (
        sigmav * rxn_energy * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_neut: NDArray[float64] = (
        sigmav * rxn_energy_neut * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_charged: NDArray[float64] = (
        sigmav * rxn_energy_charged * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )

    # Units: [MJ], [MW/m^3]
    return rxn_energy, rxn_energy_neut, rxn_energy_charged, number_power_dens, number_power_dens_neut, number_power_dens_charged


def reaction_energy_pB11(
    sigmav: NDArray[float64], heavier_fuel_species_fraction: float
) -> tuple[float, float, float, NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Proton (hydrogen)-Boron-11 reaction.

    Calculate reaction energies and power density values.

    Args:
        sigmav: :math:`\langle \sigma v \rangle` product in cm^3/s.
        heavier_fuel_species_fraction: n_heavier / (n_heavier + n_lighter) number fraction.

    Returns:
        Tuple of reaction energies and corresponding power densities.
    """
    rxn_energy: float = 8.7 * constants.value("electron volt")  # MJoules
    rxn_energy_neut: float = 0.0
    rxn_energy_charged: float = rxn_energy
    # This is accurate to within 1%
    convert_volume: float = 1e-6  # m^3/cm^3
    number_power_dens: NDArray[float64] = (
        sigmav * rxn_energy * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_neut: NDArray[float64] = (
        sigmav * rxn_energy_neut * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )
    number_power_dens_charged: NDArray[float64] = (
        sigmav * rxn_energy_charged * convert_volume * heavier_fuel_species_fraction * (1.0 - heavier_fuel_species_fraction)
    )

    # Units: [MJ], [MW/m^3]
    return rxn_energy, rxn_energy_neut, rxn_energy_charged, number_power_dens, number_power_dens_neut, number_power_dens_charged



"""Calculate the power radiated from the confined region due to the fuel and impurity species."""
import xarray as xr

from .. import formulas, named_options
from ..atomic_data import read_atomic_data
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = ["P_radiation"]


def run_calc_core_radiated_power(
    rho: Unitfull,
    electron_density_profile: Unitfull,
    electron_temp_profile: Unitfull,
    z_effective: Unitfull,
    plasma_volume: Unitfull,
    major_radius: Unitfull,
    minor_radius: Unitfull,
    magnetic_field_on_axis: Unitfull,
    separatrix_elongation: Unitfull,
    radiated_power_method: named_options.RadiationMethod,
    radiated_power_scalar: Unitfull,
    impurities: xr.DataArray,
) -> dict[str, Unitfull]:
    """Calculate the power radiated from the confined region due to the fuel and impurity species.

    Args:
        rho: :term:`glossary link<rho>`
        electron_density_profile: :term:`glossary link<electron_density_profile>`
        electron_temp_profile: :term:`glossary link<electron_temp_profile>`
        z_effective: :term:`glossary link<z_effective>`
        plasma_volume: :term:`glossary link<plasma_volume>`
        major_radius: :term:`glossary link<major_radius>`
        minor_radius: :term:`glossary link<minor_radius>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        separatrix_elongation: :term:`glossary link<separatrix_elongation>`
        radiated_power_method: :term:`glossary link<radiated_power_method>`
        radiated_power_scalar: :term:`glossary link<radiated_power_scalar>`
        impurities: :term:`glossary link<impurities>`

    Returns:
        :term:`P_radiation`

    """
    P_rad_bremsstrahlung = formulas.calc_bremsstrahlung_radiation(
        rho, electron_density_profile, electron_temp_profile, z_effective, plasma_volume
    )
    P_rad_bremsstrahlung_from_hydrogen = formulas.calc_bremsstrahlung_radiation(
        rho, electron_density_profile, electron_temp_profile, 1.0, plasma_volume
    )
    P_rad_synchrotron = formulas.calc_synchrotron_radiation(
        rho,
        electron_density_profile,
        electron_temp_profile,
        major_radius,
        minor_radius,
        magnetic_field_on_axis,
        separatrix_elongation,
        plasma_volume,
    )

    # Calculate radiated power due to Bremsstrahlung, Synchrotron and impurities
    if radiated_power_method == named_options.RadiationMethod.Inherent:
        P_radiation = radiated_power_scalar * (P_rad_bremsstrahlung + P_rad_synchrotron)
    else:
        atomic_data = read_atomic_data()

        P_rad_impurity = formulas.calc_impurity_radiated_power(
            radiated_power_method=radiated_power_method,
            rho=rho,
            electron_temp_profile=electron_temp_profile,
            electron_density_profile=electron_density_profile,
            impurities=impurities,
            plasma_volume=plasma_volume,
            atomic_data=atomic_data,
        )

        P_radiation = radiated_power_scalar * (
            P_rad_bremsstrahlung_from_hydrogen + P_rad_synchrotron + P_rad_impurity.sum(dim="dim_species")
        )

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_core_radiated_power = Algorithm(
    function=run_calc_core_radiated_power,
    return_keys=RETURN_KEYS,
)



"""Realistic functional forms for T and n for POPCON analysis.

private communication, P. Rodriguez-Fernandez (MIT PSFC), 2020

Description:
    This functional form imposes:
        1) tanh pedestal for T and n.
        2) Linear aLT profile from 0 at rho=0 to X at rho=x_a,
            where X is the specified core aLT value (default 2.0)
            and x_a is calculated by matching specified temperature_peaking (peaking)
        3) Flat aLT profile from rho=x_a to rho=1-width_ped, where
            width_ped is the pedestal width (default 0.05).
        4) T and n share the same x_a, and aLn is calculated by matching
            specified nu_n (peaking)
        5) Pedestal is rescaled to match specified volume averages for
            T and n.

Notes:
    - Not all combinations of aLT and temperature_peaking are valid. If aLT is too low,
        temperature_peaking cannot be excessively high and viceversa. The code will not
        crash, but will give profiles that do not match the specified
        temperature    peaking.
        e.g. aLT = 2.0 requires temperature_peaking to be within [1.5,3.0]

    - It is not recommended to change width_ped from the default value,
        since the Look-Up-Table hard-coded was computed using
        width_ped=0.05

    - If rho-grid is passed as argument, it is recommended to have equally
        spaced 100 points.

Example use:

    T_avol = 7.6
    n_avol = 3.1
    temperature_peaking   = 2.5
    nu_n   = 1.4

    x, T, n = evaluate_density_and_temperature_profile_fits( T_avol, n_avol, temperature_peaking, nu_n, aLT = 2.0 )

    Optionally:
        - rho-grid can be passed (100 points recommended)
        - Pedestal width can be passed (0.05 recommended)

____________________________________________________________________
"""  # TODO: figure out valid regions of fits and print a warning when they are exceeded

import warnings
from functools import cache
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
import yaml
from numpy.typing import NDArray
from scipy.interpolate import RectBivariateSpline  # type: ignore[import]

plasma_profiles_directory = Path(__file__).parent


def load_dataframe(dataset: str, df_name: str) -> pd.DataFrame:
    """Load specified dataframe for given dataset."""
    filepath = plasma_profiles_directory / dataset / f"{df_name}.csv"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df = pd.read_csv(filepath, index_col=[0, 1], header=[0, 1])

    return df


@cache
def get_df_interpolator(dataset: str, df_name: str) -> RectBivariateSpline:
    """Return an interpolator for the given dataframe of the specified dataset."""
    df = load_dataframe(dataset, df_name)
    interpolator = RectBivariateSpline(
        [np.float64(x[1]) for x in df.columns.values],
        [np.float64(x[1]) for x in df.index.values],
        df.T.values,
    )
    return interpolator


def evaluate_density_and_temperature_profile_fits(
    T_avol: float,
    n_avol: float,
    temperature_peaking: float,
    nu_n: float,
    aLT: float = 2.0,
    width_ped: float = 0.05,
    rho: Optional[NDArray[np.float64]] = None,
    dataset: str = "PRF",
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:  # TODO: fill out docstring
    """Evaluate temperature-density profile fits."""
    # ---- Get interpolator functions corresponding to this dataset
    width_interpolator = get_df_interpolator(dataset=dataset, df_name="width")
    aLT_interpolator = get_df_interpolator(dataset=dataset, df_name="aLT")

    # ---- Find parameters consistent with peaking
    x_a = width_interpolator(aLT, temperature_peaking)[0]
    aLn = aLT_interpolator(x_a, nu_n)[0]

    # ---- Evaluate profiles
    x, T, _ = evaluate_profile(T_avol, width_ped=width_ped, aLT_core=aLT, width_axis=x_a, rho=rho)
    x, n, _ = evaluate_profile(n_avol, width_ped=width_ped, aLT_core=aLn, width_axis=x_a, rho=rho)

    return x, T, n


def evaluate_profile(
    Tavol: float,
    aLT_core: float,
    width_axis: float,
    width_ped: float = 0.05,
    rho: Optional[NDArray[np.float64]] = None,
) -> tuple[NDArray[np.float64], NDArray[np.float64], float]:
    r"""This function generates a profile from :math:`\langle T \rangle`, aLT and :math:`x_a`.

    Example:
            x, T, temperature_peaking = evaluate_profile(7.6, 2.0, 0.2)
    """
    # ~~~~ Grid
    if rho is None:
        x = np.linspace(0, 1, 100)
    else:
        x = rho

    ix_c = np.argmin(np.abs(x - (1 - width_ped)))  # Extend of core
    ix_a = np.min([ix_c, np.argmin(np.abs(x - width_axis))])  # Extend of axis

    # ~~~~ aLT must be different from zero, adding non-rational small offset
    aLT_core = aLT_core + np.pi * 1e-8

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Functional Forms (normalized to pedestal temperature)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~

    # ~~~~ Pedestal
    # Notes:
    #         - Because width_ped and Teped in my function represent the top values, I need to rescale them
    #         - The tanh does not result exactly in the top value (since it's an asymptote), so I need to correct for it

    wped_tanh = width_ped / 1.5  # The pedestal width in the tanh formula is 50% inside the pedestal-top width
    Tedge_aux = 1 / 2 * (1 + np.tanh((1 - x - (wped_tanh / 2)) / (wped_tanh / 2)))
    Tedge = Tedge_aux[int(ix_c) :] / Tedge_aux[ix_c]

    # ~~~~ Core
    Tcore_aux = np.e ** (aLT_core * (1 - width_ped - x))
    Tcore = Tcore_aux[ix_a : int(ix_c)]

    # ~~~~ Axis
    Taxis_aux = np.e ** (aLT_core * (-1 / 2 * x**2 / width_axis - 1 / 2 * width_axis + 1 - width_ped))
    Taxis = Taxis_aux[:ix_a]

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Analytical Integral ("pre-factor")
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Pedestal contribution (solved with Matematica)
    I1 = -0.0277778 * width_ped * (-23.3473 + 14.6132 * width_ped)

    # Core and axis contributions
    I23 = (
        1
        / aLT_core**2
        * (
            (width_axis * aLT_core * np.e ** (width_axis * aLT_core / 2) + 1) * np.e ** (-aLT_core * (width_axis + width_ped - 1))
            + aLT_core * width_ped
            - aLT_core
            - 1
        )
    )

    # Total (this is the factor that relates Teped to Tavol)
    I = 2 * (I1 + I23)  # noqa: E741

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Evaluation of the profile
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Teped = Tavol / I
    T: NDArray[np.float64] = Teped * np.hstack((Taxis, Tcore, Tedge)).ravel()
    if np.isclose(Tavol, 0.0) and np.isclose(T[0], 0.0):
        peaking = 0.0
    else:
        peaking = float(T[0] / Tavol)

    return x, T, peaking


def load_metadata(dataset: str) -> dict[str, str]:
    r"""Load dataset metadata from YAML file.

    Args:
        dataset: name of subfolder that holds metadata.yaml

    Returns:
         metadata
    """
    filepath = plasma_profiles_directory / dataset / "metadata.yaml"
    with open(filepath) as f:
        metadata: dict[str, str] = yaml.safe_load(f)
    return metadata


def get_datasets() -> list[str]:
    """Get a list of names of valid datasets.

    Every immediate subdirectory of the source folder represents a dataset

    Returns:
         [str]*N, list of names of valid datasets
    """
    datasets = [f.name for f in plasma_profiles_directory.iterdir() if (f.is_dir() and not f.name.startswith("_"))]

    return datasets



"""Uses pint and xarray to enable unit-handling over multi-dimensional arrays."""
from typing import Union

import xarray as xr
from pint import DimensionalityError, UnitStrippedWarning

from .decorator import wraps_ufunc
from .default_units import convert_to_default_units, default_unit, magnitude_in_default_units, set_default_units
from .setup_unit_handling import Quantity, Unit, convert_units, dimensionless_magnitude, magnitude, ureg

Unitfull = Union[Quantity, xr.DataArray]

__all__ = [
    "ureg",
    "Quantity",
    "Unit",
    "Unitfull",
    "wraps_ufunc",
    "magnitude_in_default_units",
    "set_default_units",
    "default_unit",
    "convert_to_default_units",
    "convert_units",
    "magnitude",
    "dimensionless_magnitude",
    "DimensionalityError",
    "UnitStrippedWarning",
]



"""Algorithms constructed by combining several smaller algorithms."""
from .algorithm_class import CompositeAlgorithm
from .beta import calc_beta
from .core_radiated_power import calc_core_radiated_power
from .extrinsic_core_radiator import calc_extrinsic_core_radiator
from .fusion_gain import calc_fusion_gain
from .geometry import calc_geometry
from .heat_exhaust import calc_heat_exhaust
from .ohmic_power import calc_ohmic_power
from .peaked_profiles import calc_peaked_profiles
from .power_balance_from_tau_e import calc_power_balance_from_tau_e
from .q_star_from_plasma_current import calc_q_star_from_plasma_current
from .single_functions import (
    calc_auxillary_power,
    calc_average_ion_temp,
    calc_average_total_pressure,
    calc_bootstrap_fraction,
    calc_confinement_transition_threshold_power,
    calc_current_relaxation_time,
    calc_f_rad_core,
    calc_fuel_average_mass_number,
    calc_greenwald_fraction,
    calc_normalised_collisionality,
    calc_P_SOL,
    calc_peak_pressure,
    calc_ratio_P_LH,
    calc_rho_star,
    calc_triple_product,
    require_P_rad_less_than_P_in,
)
from .two_point_model_fixed_tet import two_point_model_fixed_tet
from .zeff_and_dilution_from_impurities import calc_zeff_and_dilution_from_impurities

predictive_popcon = CompositeAlgorithm(
    [
        calc_geometry,
        calc_q_star_from_plasma_current,
        calc_fuel_average_mass_number,
        calc_average_ion_temp,
        calc_zeff_and_dilution_from_impurities,
        calc_power_balance_from_tau_e,
        calc_beta,
        calc_peaked_profiles,
        calc_core_radiated_power,
        require_P_rad_less_than_P_in,
        calc_extrinsic_core_radiator,
        calc_peaked_profiles,
        calc_fusion_gain,
        calc_bootstrap_fraction,
        calc_ohmic_power,
        calc_auxillary_power,
        calc_P_SOL,
        calc_average_total_pressure,
        calc_heat_exhaust,
        two_point_model_fixed_tet,
        calc_greenwald_fraction,
        calc_confinement_transition_threshold_power,
        calc_ratio_P_LH,
        calc_f_rad_core,
        calc_normalised_collisionality,
        calc_rho_star,
        calc_triple_product,
        calc_peak_pressure,
        calc_current_relaxation_time,
    ],
    name="predictive_popcon",
)



"""Calculate the impact of core impurities on z_effective and dilution."""
import xarray as xr

from .. import formulas
from ..atomic_data import read_atomic_data
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "impurity_charge_state",
    "z_effective",
    "dilution",
    "summed_impurity_density",
    "average_ion_density",
]


def run_calc_zeff_and_dilution_from_impurities(
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    impurities: xr.DataArray,
) -> dict[str, Unitfull]:
    """Calculate the impact of core impurities on z_effective and dilution.

    Args:
        average_electron_density: :term:`glossary link<average_electron_density>`
        average_electron_temp: :term:`glossary link<average_electron_temp>`
        impurities: :term:`glossary link<impurities>`

    Returns:
        :term:`impurity_charge_state`, :term:`z_effective`, :term:`dilution`, :term:`summed_impurity_density`, :term:`average_ion_density`

    """
    starting_zeff = 1.0
    starting_dilution = 1.0

    atomic_data = read_atomic_data()

    impurity_charge_state = formulas.calc_impurity_charge_state(
        average_electron_density, average_electron_temp, impurities.dim_species, atomic_data
    )
    change_in_zeff = formulas.calc_change_in_zeff(impurity_charge_state, impurities)
    change_in_dilution = formulas.calc_change_in_dilution(impurity_charge_state, impurities)

    z_effective = starting_zeff + change_in_zeff.sum(dim="dim_species")
    dilution = starting_dilution - change_in_dilution.sum(dim="dim_species")
    summed_impurity_density = impurities.sum(dim="dim_species") * average_electron_density
    average_ion_density = dilution * average_electron_density

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_zeff_and_dilution_from_impurities = Algorithm(
    function=run_calc_zeff_and_dilution_from_impurities,
    return_keys=RETURN_KEYS,
)
